# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""
Linear regression
=================

The linear regression surrogate discipline expresses the model output
as a weighted sum of the model inputs:

.. math::

    y = w_0 + w_1x_1 + w_2x_2 + ... + w_dx_d
    + \alpha \left( \lambda \|w\|_2 + (1-\lambda) \|w\|_1 \right),

where the coefficients :math:`(w_1, w_2, ..., w_d)` and the intercept
:math:`w_0` are estimated by least square regression. They are are easily
accessible via the arguments *coefficients* and *intercept*.

The penalty level :math:`\alpha` is a non-negative parameter intended to
prevent overfitting, while the penalty ratio :math:`\lambda\in [0, 1]`
expresses the ratio between :math:`\ell_2`- and :math:`\ell_1`-regularization.
When :math:`\lambda=1`, there is no :math:`\ell_1`-regularization, and a Ridge
regression is performed. When :math:`\lambda=0`, there is no
:math:`\ell_2`-regularization, and a Lasso regression is performed. For
:math:`\lambda` between 0 and 1, an elastic net regression is performed.

One may also choose not to penalize the regression at all, by setting
:math:`\alpha=0`. In this case, a simple least squares regression is performed.

This concept is implemented through the :class:`.LinearRegression` class which
inherits from the :class:`.MLRegressionAlgo` class.

Dependence
----------
The linear model relies on the LinearRegression, Ridge, Lasso and ElasticNet
classes of the `scikit-learn library <https://scikit-learn.org/stable/modules/
linear_model.html>`_.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import array, repeat, zeros
from sklearn.linear_model import ElasticNet, Lasso
from sklearn.linear_model import LinearRegression as LinReg
from sklearn.linear_model import Ridge

from gemseo.core.dataset import Dataset
from gemseo.mlearning.regression.regression import MLRegressionAlgo
from gemseo.mlearning.transform.dimension_reduction.dimension_reduction import (
    DimensionReduction,
)
from gemseo.utils.data_conversion import DataConversion

standard_library.install_aliases()


from gemseo import LOGGER


class LinearRegression(MLRegressionAlgo):
    """ Linear regression """

    LIBRARY = "scikit-learn"
    ABBR = "LinReg"

    def __init__(
        self,
        data,
        transformer=None,
        input_names=None,
        output_names=None,
        fit_intercept=True,
        penalty_level=0.0,
        l2_penalty_ratio=1.0,
        **parameters
    ):
        """Constructor.

        :param data: learning dataset.
        :type data: Dataset
        :param transformer: transformation strategy for data groups.
            If None, do not transform data. Default: None.
        :type transformer: dict(str)
        :param input_names: names of the input variables.
        :type input_names: list(str)
        :param output_names: names of the output variables.
        :type output_names: list(str)
        :param fit_intercept: if True, fit intercept. Default: True.
        :type fit_intercept: bool
        :param penalty_level: penalty level greater or equal to 0.
            If 0, there is no penalty. Default: 0.
        :type penalty_level: float
        :param l2_penalty_ratio: penalty ratio related to the l2
            regularization. If 1, the penalty is the Ridge penalty. If 0,
            this is the Lasso penalty. Between 0 and 1, the penalty is the
            ElasticNet penalty. Default: None.
        :type l2_penalty_ratio: float
        """
        super(LinearRegression, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            fit_intercept=fit_intercept,
            penalty_level=penalty_level,
            l2_penalty_ratio=l2_penalty_ratio,
            **parameters
        )

        if penalty_level == 0.0:
            self.algo = LinReg(
                normalize=False, copy_X=False, fit_intercept=fit_intercept, **parameters
            )
        else:
            if l2_penalty_ratio == 1.0:
                self.algo = Ridge(
                    normalize=False,
                    copy_X=False,
                    fit_intercept=fit_intercept,
                    alpha=penalty_level,
                    **parameters
                )
            elif l2_penalty_ratio == 0.0:
                self.algo = Lasso(
                    normalize=False,
                    copy_X=False,
                    fit_intercept=fit_intercept,
                    alpha=penalty_level,
                    **parameters
                )
            else:
                self.algo = ElasticNet(
                    normalize=False,
                    copy_X=False,
                    fit_intercept=fit_intercept,
                    alpha=penalty_level,
                    l1_ratio=1 - l2_penalty_ratio,
                    **parameters
                )

    def _fit(self, input_data, output_data):
        """Fit the regression model.

        :param ndarray input_data: input data (2D).
        :param ndarray output_data: output data (2D).
        """
        self.algo.fit(input_data, output_data)

    def _predict(self, input_data):
        """Predict output for given input data.

        :param ndarray input_data: input data (2D).
        :return: output prediction (2D).
        :rtype: ndarray.
        """
        return self.algo.predict(input_data)

    def _predict_jacobian(self, input_data):
        """Predict Jacobian of the regression model for the given input data.

        :param ndarray input_data: input_data (2D).
        :return: Jacobian matrices (3D, one for each sample).
        :rtype: ndarray
        """
        n_samples = input_data.shape[0]
        return repeat(self.algo.coef_[None], n_samples, axis=0)

    @property
    def coefficients(self):
        """ Return the regression coefficients of the linear fit. """
        return self.algo.coef_

    @property
    def intercept(self):
        """ Return the regression intercepts of the linear fit. """
        if self.parameters["fit_intercept"]:
            intercept = self.algo.intercept_
        else:
            intercept = zeros(self.algo.coef_.shape[0])
        return intercept

    def get_coefficients(self, as_dict=True):
        """Return the regression coefficients of the linear fit
        as a numpy array or as a dict.

        :param bool as_dict: if True, returns coefficients as a dictionary.
            Default: True.
        """
        coefficients = self.coefficients
        if as_dict:
            if any(
                [
                    isinstance(transformer, DimensionReduction)
                    for _, transformer in self.transformer.items()
                ]
            ):
                raise ValueError(
                    "Coefficients are only representable in dict "
                    "form if the transformers do not change the "
                    "dimensions of the variables."
                )
            coefficients = self.__convert_array_to_dict(coefficients)
        return coefficients

    def get_intercept(self, as_dict=True):
        """Returns the regression intercept of the linear fit
        as a numpy array or as a dict.

        :param bool as_dict: if True, returns intercept as a dictionary.
            Default: True.
        """
        intercept = self.intercept
        if as_dict:
            if Dataset.OUTPUT_GROUP in self.transformer:
                raise ValueError(
                    "Intercept is only representable in dict "
                    "form if the transformers do not change the "
                    "dimensions of the output variables."
                )
            varsizes = self.learning_set.sizes
            intercept = DataConversion.array_to_dict(
                intercept, self.output_names, varsizes
            )
            intercept = {key: list(val) for key, val in intercept.items()}
        return intercept

    def __convert_array_to_dict(self, data):
        """Convert a data array into a dictionary.

        :param ndarray data: data.
        """
        varsizes = self.learning_set.sizes
        data = [
            DataConversion.array_to_dict(row, self.input_names, varsizes)
            for row in data
        ]
        data = [{key: list(val) for key, val in element.items()} for element in data]
        data = DataConversion.array_to_dict(array(data), self.output_names, varsizes)
        data = {key: list(val) for key, val in data.items()}
        return data



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Empirical estimation of statistics from a dataset
=================================================

Overview
--------

The :class:`.EmpiricalStatistics` class inherits from the
abstract :class:`.Statistics` class and aims to estimate statistics
from a :class:`.Dataset`, based on empirical estimators.

Construction
------------

A :class:`.EmpiricalStatistics` is built from a :class:`.Dataset` and
optionally a list of variables names.
In this case, statistics are only computed for these variables.
Otherwise, statistics are computed for all variables.
Lastly, the user can name its :class:`.Statistics`. By default,
the name is the concatenation of 'EmpiricalStatistics' and
and the name of the :class:`.Dataset`.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import all as npall
from numpy import max as npmax
from numpy import mean
from numpy import min as npmin
from numpy import quantile, std, var
from scipy.stats import moment

from gemseo.uncertainty.statistics.statistics import Statistics

standard_library.install_aliases()

from gemseo import LOGGER


class EmpiricalStatistics(Statistics):
    """ Empirical estimation of statistics. """

    def __init__(self, dataset, variables_names=None, name=None):
        """Constructor

        :param Dataset dataset: dataset
        :param list(str) variables_names: list of variables names
            or list of variables names. If None, the method considers
            all variables from dataset. Default: None.
        :param str name: name of the object.
            If None, use the concatenation of class and dataset names.
            Default: None."""
        name = name or dataset.name
        super(EmpiricalStatistics, self).__init__(dataset, variables_names, name)

    def maximum(self):
        """Get the maximum.

        :return: maximum
        :rtype: dict
        """
        result = {name: npmax(self.dataset[name], 0) for name in self.names}
        return result

    def mean(self):
        """Get the mean.

        :return: mean
        :rtype: dict
        """
        result = {name: mean(self.dataset[name], 0) for name in self.names}
        return result

    def minimum(self):
        """Get the minimum.

        :return: minimum
        :rtype: dict
        """
        result = {name: npmin(self.dataset[name], 0) for name in self.names}
        return result

    def probability(self, thresh, greater=True):
        """Compute a probability associated to a threshold. This threshold
        is a dictionary of arrays indexed by variables names.
        For a multidimensional variable, the probability to be greater
        (or lower) than the threshold is defined as the probability that all
        variables components are greater (respectively lower) than their
        counterparts in the threshold.

        :param dict thresh: threshold
        :param bool greater: if True, compute the probability the probability
            of exceeding the threshold, if False, compute the reverse.
            Default: True.
        :return: probability
        """
        if greater:
            result = {
                name: mean(npall(self.dataset[name] >= thresh[name], 1))
                for name in self.names
            }
        else:
            result = {
                name: mean(npall(self.dataset[name] <= thresh[name], 1))
                for name in self.names
            }
        return result

    def quantile(self, prob):
        """Get the quantile associated to a given probability.

        :param int merge: if True, merge variables. Default: True.
        :return: quantile
        :rtype: dict
        """
        result = {name: quantile(self.dataset[name], prob, 0) for name in self.names}
        return result

    def standard_deviation(self):
        """Get the standard deviation.

        :return: standard deviation
        :rtype: dict
        """
        result = {name: std(self.dataset[name], 0) for name in self.names}
        return result

    def variance(self):
        """Get the variance.

        :return: variance
        :rtype: dict
        """
        result = {name: var(self.dataset[name], 0) for name in self.names}
        return result

    def moment(self, order):
        """Compute the central moment for a given order.

        :param int order: moment order.
        :return: moment
        :rtype: dict
        """
        result = {name: moment(self.dataset[name], order, 0) for name in self.names}
        return result

    def range(self):
        """Get the range of variables.

        :return: range of variables
        """
        lower = self.minimum()
        upper = self.maximum()
        result = {name: upper[name] - lower[name] for name in self.names}
        return result



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Estimation of statistics from a dataset
=======================================

Overview
--------

The abstract :class:`.Statistics` class implements the concept of
statistics library. It is enriched by concrete classes
such as :class:`.EmpiricalStatistics` and :class:`.ParametricStatistics`.

Construction
------------

A :class:`.Statistics` is built from a :class:`.Dataset` and optionally
a list of variables names. In this case, statistics are only computed
for these variables. Otherwise, statistics are computed for all variables.
Lastly, the user can name its :class:`.Statistics`. By default,
the name is the concatenation of the name of the class
overloading :class:`.Statistics`
and the name of the :class:`.Dataset`.

Capabilities
------------

A :class:`.Statistics` returns standard descriptive and statistical measures
for the different variables:

- :meth:`.Statistics.minimum`: the minimum value,
- :meth:`.Statistics.maximum`: the maximum value,
- :meth:`.Statistics.range`: the difference between minimum and maximum values,
- :meth:`.Statistics.mean`: the expectation, a.k.a. mean value,
- :meth:`.Statistics.moment`: the central moment which is a the expected value
  of a specified integer power of the deviation from the mean,
- :meth:`.Statistics.variance`: the variance, which is the mean squared
  variation around the mean value,
- :meth:`.Statistics.standard_deviation`: the standard deviation, which is the
  square root of the variance,
- :meth:`.Statistics.quantile`: the quantile associated with a probability,
  which is the cut point diving the range into a first continuous interval
  with this given probability and a second continuous interval
  with the complementary probability; common *q*-quantiles dividing
  the range into *q* continuous interval with equal probabilities are also
  implemented:

    - :meth:`.Statistics.median` which implements the 2-quantile (50%).
    - :meth:`.Statistics.quartile` whose order (1, 2 or 3) implements
      the 4-quantiles (respectively 25%, 50% and 75%),
    - :meth:`.Statistics.percentile` whose order (1, 2, ..., 99) implements
      the 100-quantiles (1%, 2%, ..., 99%),

- :meth:`.Statistics.probability`: the probability that the random variable
  is larger or smaller than a certain threshold,
- :meth:`.Statistics.tolerance_interval`: the left-sided, right-sided or
  both-sided tolerance interval associated with a given coverage level
  and a given confidence level, which is
  a statistical interval within which, with some confidence level,
  a specified proportion of the random variable realizations falls
  (this proportion is the coverage level)

    - :meth:`.Statistics.a_value`: the A-value, which is the lower bound of the
      left-sided tolerance interval associated with a coverage level
      equal to 99% and a confidence level equal to 95%,
    - :meth:`.Statistics.b_value`: the B-value, which is the lower bound of the
      left-sided tolerance interval associated with a coverage level equal
      to 90% and a confidence level equal to 95%,

"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library

standard_library.install_aliases()

from gemseo import LOGGER


class Statistics(object):
    """Abstract class for Statistics library interface."""

    def __init__(self, dataset, variables_names=None, name=None):
        """
        Constructor

        :param Dataset dataset: dataset
        :param list(str) variables_names: list of variables names.
            If None, the method considers
            all variables from loaded dataset. Default: None.
        :param str name: name of the object.
            If None, use the concatenation of class and dataset names.
            Default: None.
        """
        default_name = self.__class__.__name__ + "_" + dataset.name
        self.name = name or default_name
        msg = 'Create "' + self.name + '"'
        msg += ", a " + self.__class__.__name__ + " library."
        LOGGER.info(msg)
        self.dataset = dataset.get_all_data(by_group=False, as_dict=True)
        self.n_samples = dataset.n_samples
        self.names = variables_names or dataset.variables
        self.n_variables = dataset.n_variables

    def __str__(self):
        msg = self.name + "\n"
        msg += "| n_samples: " + str(self.n_samples) + "\n"
        msg += "| n_variables: " + str(self.n_variables) + "\n"
        msg += "| variables: " + str(self.names) + "\n"
        return msg

    def tolerance_interval(self, coverage, confidence=0.95, side="both"):
        """Compute the tolerance interval (TI) for a given minimum percentage
        of the population and a given confidence level.

        :param float coverage: minimum percentage of belonging to the TI.
        :param float confidence: level of confidence in [0,1]. Default: 0.95.
        :param str side: kind of interval: 'lower' for lower-sided TI,
            'upper' for upper-sided TI and 'both for both-sided TI.
        :return: tolerance limits
        """
        raise NotImplementedError

    def a_value(self):
        """Compute the b-value.

        :return: b-value
        """
        result = self.tolerance_interval(1 - 0.1, 0.99, "lower")
        result = {name: value[0] for name, value in result.items()}
        return result

    def b_value(self):
        """Compute the b-value.

        :return: b-value
        """
        result = self.tolerance_interval(1 - 0.1, 0.95, "lower")
        result = {name: value[0] for name, value in result.items()}
        return result

    def maximum(self):
        """Compute the maximum.

        :return: maximum
        """
        raise NotImplementedError

    def mean(self):
        """Compute the mean.

        :return: mean
        """
        raise NotImplementedError

    def minimum(self):
        """Compute the minimum.

        :return: minimum
        """
        raise NotImplementedError

    def median(self):
        """Compute the median.

        :param options: options
        :return: median
        """
        result = self.quantile(0.5)
        return result

    def percentile(self, order):
        """Compute the percentile.

        :param int order: percentile order, e.g. 4.
        :return: percentile
        """
        if not isinstance(order, int) or order > 100 or order < 0:
            raise TypeError(
                "Percentile order must be an integer " "between 0 and 100 inclusive."
            )
        prob = order / 100.0
        result = self.quantile(prob)
        return result

    def probability(self, thresh, greater):
        """Compute a probability associated to a threshold.

        :param float thresh: threshold
        :param bool greater: if True, compute the probability the probability
            of exceeding the threshold, if False, compute the reverse.
        :return: probability
        """
        raise NotImplementedError

    def quantile(self, prob):
        """Compute a quantile associated to a probability.

        :param float prob: probability between 0 and 1
        :return: quantile
        """
        raise NotImplementedError

    def quartile(self, order):
        """Compute a quartile.

        :param int order: quartile order in [1,2,3]
        :return: quartile
        """
        quartiles = [0.25, 0.5, 0.75]
        if order not in [1, 2, 3]:
            raise ValueError("Quartile order must be in [1,2,3]")
        prob = quartiles[order - 1]
        result = self.quantile(prob)
        return result

    def range(self):
        """Compute the range

        :return: range
        """
        raise NotImplementedError

    def standard_deviation(self):
        """Compute a standard_deviation.

        :return: standard deviation
        """
        raise NotImplementedError

    def variance(self):
        """Compute a variance.

        :return: variance
        """
        raise NotImplementedError

    def moment(self, order):
        """Compute the moment for a given order.

        :param int order: moment index
        :return: moment
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Parse source code to extract information
****************************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import re
from builtins import range, str
from inspect import getargspec, getdoc

from future import standard_library

standard_library.install_aliases()


class SourceParsing(object):
    """
    Parse source code to extract information
    """

    @staticmethod
    def get_options_doc(method):
        """
        Get the documentation of a method

        :param method: the method to retreive the doc from
        :returns: the dictionary of options meaning
        """
        doc = getdoc(method)
        if doc is None:
            raise ValueError("Empty doc for " + str(method))
        pattern = ":param ([\*\w]+): (.*?)"  # pylint: disable=W1401
        pattern += "(?:(?=:param)|(?=:return)|\Z)"  # pylint: disable=W1401
        param_re = re.compile(pattern, re.S)
        doc_list = param_re.findall(doc)
        return {txt[0]: txt[1].replace(" " * 4, "") for txt in doc_list}

    @staticmethod
    def get_default_options_values(klass):
        """
        Get the options default values for the given class
        Only addresses kwargs

        :param name : name of the class
        :returns: the dict option name: option default value
        """
        args, _, _, defaults = getargspec(klass.__init__)
        if "self" in args:
            args.remove("self")
        n_def = len(defaults)

        args_dict = {args[-n_def:][i]: defaults[i] for i in range(n_def)}

        return args_dict



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A factory to execute DOE algorithms from their class names
**********************************************************
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library

from gemseo.algos.doe.doe_lib import DOELibrary
from gemseo.algos.driver_factory import DriverFactory

standard_library.install_aliases()


class DOEFactory(DriverFactory):
    """DOE factory to create DOE libraries, see DriverFactory"""

    def __init__(self):
        """Test and import optimization library if it is available."""
        super(DOEFactory, self).__init__(DOELibrary, "gemseo.algos.doe")



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Calibration of a machine learning algorithm
===========================================

A machine learning algorithm depends on hyperparameters,
e.g. number of clusters for a clustering algorithm,
regularization constant for a regression model,
kernel for a Gaussian process regression, ...
Its quality of generalization depends on the values of these hyperparameters.
Thus, the hyperparameters minimizing the learning quality measure are rarely
those minimizing the generalization one. Classically, the generalization one
decreases before growing again as the model becomes more complex,
while the learning error keeps decreasing. This phenomenon is called the
curse of dimensionality.

In this module, the :class:`.MLAlgoCalibration` class aims to calibrate the
hyperparameters in order to minimize this generalization quality measure
over a calibration parameter space. This class relies on the
:class:`.MLAlgoAssessor` class which is a discipline
(:class:`.MDODiscipline`)
built from a machine learning algorithm (:class:`.MLAlgo`),
a dataset (:class:`.Dataset`), a quality measure (:class:`.MLQualityMeasure`)
and various options for data scaling, quality measure
and machine learning algorithm. The inputs of this discipline are
hyperparameters of the machine learning algorithm while the output is
the quality criterion.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import argmin, array

from gemseo.core.discipline import MDODiscipline
from gemseo.core.doe_scenario import DOEScenario
from gemseo.core.mdo_scenario import MDOScenario
from gemseo.mlearning.core.factory import MLAlgoFactory

standard_library.install_aliases()


class MLAlgoAssessor(MDODiscipline):
    """ Discipline assessing the quality of a machine learning algorithm """

    def __init__(
        self,
        algo,
        dataset,
        parameters,
        measure,
        measure_options=None,
        transformer=None,
        **algo_options
    ):
        """Constructor

        :param str algo: machine learning algorithm name.
        :param Dataset dataset: learning dataset.
        :param list(str) parameters: parameters.
        :param MLQualityMeasure measure: quality measure.
        :param dict measure_options: options of the quality measures.
        :param dict(str) transformer: transformation strategy for data groups.
            If None, do not transform data. Default: None.
        :param algo_options: options of the machine learning algorithm.
        """
        super(MLAlgoAssessor, self).__init__()
        self.input_grammar.initialize_from_data_names(parameters)
        self.output_grammar.initialize_from_data_names(["criterion", "learning"])
        self.algo = algo
        self.measure = measure
        self.measure_options = measure_options or {}
        self.parameters = algo_options
        self.data = dataset
        self.transformer = transformer
        self.algos = []

    def _run(self):
        """ run method. """
        inputs = self.get_input_data()
        for index in inputs:
            if len(inputs[index]) == 1:
                inputs[index] = inputs[index][0]
        self.parameters.update(inputs)
        factory = MLAlgoFactory()
        algo = factory.create(
            self.algo, data=self.data, transformer=self.transformer, **self.parameters
        )
        algo.learn()
        measure = self.measure(algo)
        learning = measure.evaluate(multioutput=False)
        criterion = measure.evaluate(multioutput=False, **self.measure_options)
        self.store_local_data(criterion=array([criterion]), learning=array([learning]))
        self.algos.append(algo)


class MLAlgoCalibration(object):
    """ Calibration of a machine learning algorithm """

    CRITERION = "criterion"

    def __init__(
        self,
        algo,
        dataset,
        parameters,
        calibration_space,
        measure,
        measure_options=None,
        transformer=None,
        use_doe=True,
        **algo_options
    ):
        """Constructor

        :param str algo: machine learning algorithm name.
        :param Dataset dataset: learning dataset.
        :param list(str) parameters: parameters.
        :param DesignSpace calibration_space: calibration space.
        :param MLQualityMeasure measure: quality measure.
        :param dict measure_options: options of the quality measures.
        :param dict(str) transformer: transformation strategy for data groups.
            If None, do not transform data. Default: None.
        :param bool use_doe: if True, use a DOEScenario to calibrate
            the ML algorithm. Otherwise, use a MDOScenario. Default: True.
        :param algo_options: options of the machine learning algorithm.
        """
        disc = MLAlgoAssessor(
            algo,
            dataset,
            parameters,
            measure,
            measure_options,
            transformer,
            **algo_options
        )
        disc.set_cache_policy(disc.MEMORY_FULL_CACHE)
        if use_doe:
            self.scenario = DOEScenario(
                [disc], "DisciplinaryOpt", self.CRITERION, calibration_space
            )
        else:
            self.scenario = MDOScenario(
                [disc], "DisciplinaryOpt", self.CRITERION, calibration_space
            )
        self.dataset = None
        self.optimal_parameters = None
        self.optimal_criterion = None
        self.optimal_algorithm = None

    def execute(self, input_data):
        """Execute the calibration from optimization or DOE data.

        :param dict input_data: optimization or DOE data.
        :return: optimal hyperparameters, optimal criterion.
        :rtype: dict, ndarray
        """
        self.scenario.disciplines[0].cache.clear()
        self.scenario.execute(input_data)
        x_opt = self.scenario.design_space.get_current_x_dict()
        f_opt = self.scenario.get_optimum().f_opt
        cache = self.scenario.disciplines[0].cache
        self.dataset = cache.export_to_dataset(by_group=False)
        algo_opt = self.algos[argmin(self.get_history(self.CRITERION))]
        self.optimal_parameters = x_opt
        self.optimal_criterion = f_opt
        self.optimal_algorithm = algo_opt

    def get_history(self, name):
        """Get history of a given variable.

        :param str name: variable name.
        :return: history of the variable.
        :rtype: ndarray
        """
        if self.dataset is not None:
            data = self.dataset.data[name]
        return data

    @property
    def algos(self):
        """ List of trained algorithms. """
        return self.scenario.disciplines[0].algos



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Kernel Principal Component Analysis
===================================

The :class:`KPCA` class implements the KCPA wraps the KPCA from Scikit-learn.

Dependence
----------
This dimension reduction algorithm relies on the PCA class
of the `scikit-learn library <https://scikit-learn.org/stable/modules/
generated/sklearn.decomposition.PCA.html>`_.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from sklearn.decomposition import KernelPCA

from gemseo.mlearning.transform.dimension_reduction.dimension_reduction import (
    DimensionReduction,
)

standard_library.install_aliases()


class KPCA(DimensionReduction):
    """ Kernel principal component dimension reduction algorithm. """

    def __init__(
        self,
        name="KPCA",
        n_components=5,
        fit_inverse_transform=True,
        kernel="linear",
        **parameters
    ):
        """Constructor.

        :param str name: transformer name. Default: 'KPCA'.
        :param int n_components: number of components. Default: 5.
        :param bool fit_inverse_transform: Learn the inverse transform for
            non-precomputed kernels. Default: True.
        :param str kernel: kernel name ('linear', 'poly', 'rbf',
            'sigmoid', 'cosine' or 'precomputed'). Default: 'linear'.
        :param parameters: Optional parameters for sklearn KPCA constructor.
        """
        super(KPCA, self).__init__(name, n_components=n_components, **parameters)
        self.algo = KernelPCA(
            n_components,
            fit_inverse_transform=fit_inverse_transform,
            kernel=kernel,
            **parameters
        )

    def fit(self, data):
        """Fit transformer to data.

        :param ndarray data: data to be fitted.
        """
        self.algo.fit(data)

    def transform(self, data):
        """Transform data.

        :param ndarray data: data  to be transformed.
        :return: transformed data.
        :rtype: ndarray
        """
        return self.algo.transform(data)

    def inverse_transform(self, data):
        """Perform an inverse transform on the data.

        :param ndarray data: data to be inverse transformed.
        :return: inverse transformed data.
        :rtype: ndarray
        """
        return self.algo.inverse_transform(data)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Error measure
=============

The :mod:`~gemseo.mlearning.qual_measure.error_measure` module implements
the concept of error measures for machine learning algorithms.

This concept is implemented through the :class:`.MLErrorMeasure` class
and implements the different evaluation methods.

The error measure class is adapted for supervised machine learning algorithms,
as it measures the error of a predicted value to some reference value.
"""
from __future__ import absolute_import, division, unicode_literals

import numpy.random as npr
from future import standard_library
from numpy import arange, array_split
from numpy import delete as npdelete
from numpy import vstack

from gemseo.mlearning.qual_measure.quality_measure import MLQualityMeasure

standard_library.install_aliases()


class MLErrorMeasure(MLQualityMeasure):
    """ Error measure for machine learning. """

    def evaluate_learn(self, multioutput=True):
        """Evaluate quality measure using the learning dataset.

        :param bool multioutput: if True, return the error measure for each
            output component. Otherwise, average these errors. Default: True.
        :return: quality measure value.
        """
        if not self.algo.is_trained:
            self.algo.learn()
        in_grp = self.algo.learning_set.INPUT_GROUP
        out_grp = self.algo.learning_set.OUTPUT_GROUP
        inputs = self.algo.learning_set.get_data_by_group(in_grp)
        outputs = self.algo.learning_set.get_data_by_group(out_grp)
        predictions = self.algo.predict(inputs)
        measure = self._compute_measure(outputs, predictions, multioutput)
        return measure

    def evaluate_test(self, test_data, multioutput=True):
        """Evaluate quality measure using a test dataset.

        :param Dataset test_data: test data.
        :param bool multioutput: if True, return the error measure for each
            output component. Otherwise, average these errors. Default: True.
        :return: quality measure value.
        """
        if not self.algo.is_trained:
            self.algo.learn()
        in_grp = test_data.INPUT_GROUP
        out_grp = test_data.OUTPUT_GROUP
        inputs = test_data.get_data_by_group(in_grp)
        outputs = test_data.get_data_by_group(out_grp)
        predictions = self.algo.predict(inputs)
        measure = self._compute_measure(outputs, predictions, multioutput)
        return measure

    def evaluate_kfolds(self, n_folds=5, multioutput=True):
        """Evaluate quality measure using the k-folds technique.

        :param int n_folds: number of folds. Default: 5.
        :param bool multioutput: if True, return the error measure for each
            output component. Otherwise, average these errors. Default: True.
        :return: quality measure value.
        """
        n_samples = self.algo.learning_set.n_samples
        inds = arange(n_samples)
        folds = array_split(inds, n_folds)

        in_grp = self.algo.learning_set.INPUT_GROUP
        out_grp = self.algo.learning_set.OUTPUT_GROUP
        inputs = self.algo.learning_set.get_data_by_group(in_grp)
        outputs = self.algo.learning_set.get_data_by_group(out_grp)

        expected = []
        predicted = []
        for n_fold in range(n_folds):
            fold = folds[n_fold]
            train = npdelete(inds, fold)
            self.algo.learn(samples=train)

            predicted.append(self.algo.predict(inputs[fold]))
            expected.append(outputs[fold])

        expected = vstack(expected)
        predicted = vstack(predicted)

        return self._compute_measure(expected, predicted, multioutput)

    def evaluate_bootstrap(self, n_replicates=100, multioutput=True):
        """Evaluate quality measure using the bootstrap technique.

        :param int n_replicates: number of bootstrap replicates. Default: 100.
        :param bool multioutput: if True, return the error measure for each
            output component. Otherwise, average these errors. Default: True.
        :return: quality measure value.
        """
        n_samples = self.algo.learning_set.n_samples
        inds = arange(n_samples)

        in_grp = self.algo.learning_set.INPUT_GROUP
        out_grp = self.algo.learning_set.OUTPUT_GROUP
        inputs = self.algo.learning_set.get_data_by_group(in_grp)
        outputs = self.algo.learning_set.get_data_by_group(out_grp)

        expected = []
        predicted = []
        for _ in range(n_replicates):
            train = npr.choice(n_samples, n_samples)
            test = npdelete(inds, train)
            self.algo.learn(samples=train)
            predicted.append(self.algo.predict(inputs[test]))
            expected.append(outputs[test])

        expected = vstack(expected)
        predicted = vstack(predicted)

        return self._compute_measure(expected, predicted, multioutput)

    def _compute_measure(self, outputs, predictions, multioutput=True):
        """Compute error measure.

        :param ndarray outputs: reference outputs.
        :param ndarray predictions: predicted outputs.
        :param bool multioutput: if True, return the error measure for each
            output component. Otherwise, average these errors. Default: True.
        :return: measure value.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Damien Guenot
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Quadratic approximations of functions from the optimization history
*******************************************************************
"""

from __future__ import absolute_import, division, print_function, unicode_literals

from math import ceil
from os.path import splitext

import numpy as np
import pylab
from future import standard_library
from matplotlib import pyplot
from matplotlib.colors import SymLogNorm
from matplotlib.ticker import LogFormatter
from pylab import plt

from gemseo.post.core.colormaps import PARULA
from gemseo.post.core.hessians import SR1Approx
from gemseo.post.opt_post_processor import OptPostProcessor

standard_library.install_aliases()
from gemseo import LOGGER


class QuadApprox(OptPostProcessor):
    """
    The **QuadApprox** post processing
    performs a quadratic approximation of a given function
    from an optimization history
    and plot the results as cuts of the approximation.

    The function index can be passed as option.
    It is possible either to save the plot, to show the plot or both.
    """

    SR1_APPROX = "SR1"

    def __init__(self, opt_problem):
        """
        Constructor

        :param opt_problem : the optimization problem to run
        """
        super(QuadApprox, self).__init__(opt_problem)
        self.grad_opt = None

    def _plot(
        self,
        function,
        save=True,
        show=False,
        file_path="quad_approx",
        func_index=None,
        extension="pdf",
    ):
        """
        Builds the plot and saves it

        :param function: function name to build quadratic approximation
        :type function: str
        :param show: if True, displays the plot windows
        :type show: bool
        :param save: if True, exports plot to pdf
        :type save: bool
        :param file_path: the base paths of the files to export
        :type file_path: str
        :param share_y: if True, all Y axis are the same,
            useful to compare sensitivities
        :type share_y: bool
        :param func_index: functional index
        :type func_index: int
        :param extension: file extension
        :type extension: str
        """
        b_mat = self.__build_approx(function, func_index)
        self.out_data_dict["b_mat"] = b_mat
        fig = self.__plot_hessian(b_mat, function)

        root = splitext(file_path)[0]
        file_path = root + "_hess_approx"
        self._save_and_show(
            fig, save=save, show=show, file_path=file_path, extension=extension
        )
        fig = self.__plot_variations(b_mat)

        file_path = root + "_quad_approx"
        self._save_and_show(
            fig, save=save, show=show, file_path=file_path, extension=extension
        )

    def __build_approx(self, function, func_index):
        """
        Builds the approximation

        :param method: SR1 or BFGS
        """
        apprx = SR1Approx(self.database)
        n_x = self.opt_problem.get_dimension()
        max_i = int(1.5 * n_x)
        # Avoid using alpha scaling for hessian otherwise diagonal is messy
        out = apprx.build_approximation(
            function, at_most_niter=max_i, return_x_grad=True, func_index=func_index
        )
        b_mat = out[0]
        grad_opt = out[-1]

        self.grad_opt = grad_opt
        return b_mat

    @staticmethod
    def __plot_hessian(hessian, function):
        """
        Plots the Hessian of the function

        :param hessian: the hessian
        """
        fig = pylab.figure()
        pylab.plt.xlabel(r"$x_i$", fontsize=16)
        pylab.plt.ylabel(r"$x_j$", fontsize=16)
        vmax = max(abs(np.max(hessian)), abs(np.min(hessian)))
        linthresh = 10 ** ((np.log10(vmax) - 5.0))
        # SymLog is a symetric log scale adapted to negative values
        pylab.imshow(
            hessian,
            cmap=PARULA,
            interpolation="nearest",
            norm=SymLogNorm(linthresh=linthresh, vmin=-vmax, vmax=vmax),
        )
        pylab.grid(True)
        l_f = LogFormatter(base=10, labelOnlyBase=False)
        thick_min = int(np.log10(linthresh))
        thick_max = int(np.log10(vmax))
        thick_num = thick_max - thick_min + 1
        lvls_pos = np.logspace(thick_min, thick_max, num=thick_num, base=10.0)
        levels_neg = np.sort(-lvls_pos)
        levels = np.concatenate((levels_neg, lvls_pos))
        pylab.plt.colorbar(ticks=levels, format=l_f)
        fig.suptitle("Hessian matrix SR1 approximation of " + str(function))
        return fig

    @staticmethod
    def unnormalize_vector(xn_array, ivar, lower_bounds, upper_bounds):
        """Unormalize a variable with respect to bounds

        :param xn_array: normalized variable (array)
        :param ivar: index of variable bound
        :param lower_bounds: param upper_bounds:
        :param upper_bounds:
        :returns: unnormalized array
        :rtype: numpy array

        """
        l_b = lower_bounds[ivar]
        u_b = upper_bounds[ivar]
        return (u_b - l_b) * xn_array * 0.5 + 0.5 * (l_b + u_b)

    def __plot_variations(self, hessian):
        """Plots the variation plots of the function wrt all variables """
        ndv = hessian.shape[0]
        ncols = int(np.sqrt(ndv)) + 1
        nrows = int(ceil(float(ndv) / ncols))

        xn_vars = np.arange(-1.0, 1.0, 0.01)
        lower_bounds = self.opt_problem.design_space.get_lower_bounds()
        upper_bounds = self.opt_problem.design_space.get_upper_bounds()
        fig = plt.figure()

        for i in range(ndv):
            ax_i = plt.subplot(nrows, ncols, i + 1)
            f_vals = xn_vars ** 2 * hessian[i, i] + self.grad_opt[i] * xn_vars
            self.out_data_dict[i] = f_vals

            x_vars = self.unnormalize_vector(xn_vars, i, lower_bounds, upper_bounds)
            ax_i.plot(x_vars, f_vals, "-", lw=2)

            start, stop = ax_i.get_xlim()
            ax_i.xaxis.set_ticks(np.arange(start, stop, 0.4999999 * (stop - start)))
            start, stop = ax_i.get_ylim()
            ax_i.yaxis.set_ticks(np.arange(start, stop, 0.24999999 * (stop - start)))
            ax_i.set_xlabel(r"$x_{" + str(i) + "}$", fontsize=14)
        pyplot.tight_layout()
        return fig



# -*- coding: utf-8 -*-
# https://github.com/noamraph/tqdm
# the MIT License (MIT)
# Copyright (c) 2013 noamraph
# permission is hereby granted, free of charge,
# to any person obtaining a copy of
# this software and associated documentation
# files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# the, copy, modify, merge, publish, distribute,
# sublicense, and/or sell copies of
# the Software, and to permit persons to whom the
# Software is furnished to do so,
# object to the following conditions:
# the above copyright notice and this permission
# notice shall be included in all
# copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO
# THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
A progress bar implementation
*****************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from builtins import chr, int
from multiprocessing import Value
from numbers import Number
from numpy.linalg import norm
from timeit import default_timer as timer

from future import standard_library

from gemseo.api import configure_logger


standard_library.install_aliases()


from gemseo import LOGGER

N_BARS = 10


class StatusPrinter(object):
    """
    Prints the status, recors the last printed status
    This class is multiprocessing-safe : the counters values are shared among all processes
    """

    def __init__(self):
        self.last_printed_len = 0

    def print_status(self, status):
        """
        Prints the status

        :param status: the status
        """
        LOGGER.info(status + " " * max(self.last_printed_len - len(status), 0))
        self.last_printed_len = len(status)


class Tqdm(object):
    """TQDM iterator progress bar"""

    def __init__(self, max_iter, desc="", mininterval=0.5, miniters=1):
        """
        Prints a progress meter and updates it every time
        the next() ethod is   called.
        'desc' can contain a short string,
        describing the progress, that is added
        in the beginning of the line.
        If less than mininterval seconds or  miniters iterations have passed
        since the last progress meter update, it is not updated again.
        """
        self.max_iter = max_iter
        self.desc = desc
        self.mininterval = mininterval
        self.miniters = miniters
        self.prefix = self.desc + ": " if self.desc else ""
        self.stat_printer = StatusPrinter()
        self._last_print_t = Value("d", -1.0)
        self._start_t = Value("d", -1.0)
        self._last_print_n = Value("i", -1)
        # The counter is shared among all processes when run in parallel
        self._current_n = Value("i", 0)
        self._closed = Value("b", False)

    def _print_current(self, cur_t, obj_val=None):
        """
        Prints the current status

        :param cur_t: current time (Default value = None)
        :param obj_val: curent objective value (Default value = None)
        """
        formatted = format_meter(
            self._current_n.value, self.max_iter, cur_t - self._start_t.value, obj_val
        )
        self.stat_printer.print_status(self.prefix + formatted)

    def close(self, obj_val=None):
        """
        Closes the progress bar
        """
        with self._current_n.get_lock() and self._last_print_n.get_lock() and self._closed.get_lock():
            if self._last_print_n.value != self._current_n.value:
                if self._current_n.value > self.max_iter:
                    self._current_n.value = self.max_iter
                self._print_current(timer(), obj_val)
                self._last_print_n.value = self._current_n.value
            self._closed.value = True

    def next(self, obj_val=None):
        """
        Iterates the progress bar

        :param obj_val: objective value (Default value = None)
        """

        if self._closed.value:
            raise ValueError("Progress bar is closed while calling next()")
        # Now the object was created and processed, so we can print the
        # meter.
        cur_t = timer()
        with self._current_n.get_lock() and self._last_print_n.get_lock() and self._last_print_t.get_lock() and self._start_t.get_lock():
            if self._current_n.value == self.max_iter:
                raise StopIteration
            if self._current_n.value >= self.miniters + self._last_print_n.value:
                # We check the counter first, to reduce the overhead of

                if cur_t >= self.mininterval + self._last_print_t.value:
                    if self._last_print_t.value < 0.0:
                        # First call, we need to initialize the variables
                        self._last_print_t.value = cur_t
                        self._start_t.value = cur_t
                    self._print_current(cur_t, obj_val)
                    self._last_print_n.value = self._current_n.value
                    self._last_print_t.value = cur_t

            self._current_n.value += 1


def format_interval(time_inter):
    """
    Formats a time interval

    :param time_inter: the time interval to format
    """
    mins, sec = divmod(int(time_inter), 60)
    hours, mins = divmod(mins, 60)
    if hours:
        return "%d:%02d:%02d" % (hours, mins, sec)
    else:
        return "%02d:%02d" % (mins, sec)


def format_meter(n_iter, max_iter, elapsed, obj_val=None):
    """
    Format a message

    :param n_iter: number of finished iterations
    :param obj_val:  objective value (Default value = None)
    :param max_iter: total number of iterations, or None
    :param elapsed: number of seconds passed since start
    """
    if n_iter > max_iter:
        max_iter = None

    elapsed_str = format_interval(elapsed)
    if not elapsed:
        rate_str, unit = "?", "iters/sec"
    else:
        rate, unit = get_printable_rate(n_iter, elapsed)
        rate_str = "%5.2f" % rate
    if max_iter:
        frac = float(n_iter) / max_iter

        bar_length, frac_bar_length = divmod(int(frac * N_BARS * 8), 8)
        bar_l = chr(0x2588) * bar_length
        frac_bar = chr(0x2590 - frac_bar_length) if frac_bar_length else " "

        percentage = "%3d%%" % (frac * 100)
        if bar_length < N_BARS:
            full_bar = bar_l + frac_bar + " " * max(N_BARS - bar_length - 1, 0)
        else:
            full_bar = bar_l + " " * max(N_BARS - bar_length, 0)
        left_str = (
            format_interval(elapsed / n_iter * (max_iter - n_iter)) if n_iter else "?"
        )

        if obj_val is None:
            format_str = "|%s| %d/%d %s [elapsed: %s left: %s, %s " + unit + "]"
            return format_str % (
                full_bar,
                n_iter,
                max_iter,
                percentage,
                elapsed_str,
                left_str,
                rate_str,
            )
        else:
            if obj_val is None or isinstance(obj_val, Number) or obj_val.size == 1:
                tmp = obj_val
            else:
                tmp = norm(obj_val)
            obj_s = "%5.2f" % tmp
            format_str = "|%s| %d/%d %s "
            format_str += "[elapsed: %s left: %s, %s " + unit + " obj: %s ]"
            return format_str % (
                full_bar,
                n_iter,
                max_iter,
                percentage,
                elapsed_str,
                left_str,
                rate_str,
                obj_s,
            )
    else:
        return "%d [elapsed: %s, %s " + unit + "]" % (n_iter, elapsed_str, rate_str)


def get_printable_rate(n_iter, elapsed):
    rps = float(n_iter) / elapsed
    if rps >= 0:
        rate = rps
        unit = "iters/sec"

    rpm = rps * 60.0
    if rpm < 60.0:
        rate = rpm
        unit = "iters/min"

    rph = rpm * 60.0
    if rph < 60:
        rate = rph
        unit = "iters/hour"

    rpd = rph * 24.0
    if rpd < 24:
        rate = rpd
        unit = "iters/day"

    return rate, unit



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Parametric estimation of statistics from a dataset
==================================================

Overview
--------

The :class:`.ParametricStatistics` class inherits from the
abstract :class:`.Statistics` class and aims to estimate statistics
from a :class:`.Dataset`, based on a collection of
candidate parametric distribution calibrated from this :class:`.Dataset`.
For each variable, parameters of these distributions are calibrated
from the :class:`.Dataset`
and the fitted parametric :class:`.Distribution` which is optimal
in the sense of a goodness-of-fit criterion and a selection criterion
is selected to estimate :class:`.Statistics` associated with this variable.
The :class:`.ParametricStatistics` relies on the OpenTURNS library through
the :class:`.OTDistribution` and :class:`.OTDistributionFitter` classes.

Construction
------------

The :class:`.ParametricStatistics` is built from two mandatory arguments:

- a dataset,
- a list of distributions names,

and can consider optional arguments:

- a subset of variables names (by default, statistics are computed
  for all variables),
- a fitting criterion name (by default, BIC is used;
  see :meth:`.ParametricStatistics.get_available_criteria`
  and :meth:`.ParametricStatistics.get_significance_tests`
  for more information),
- a level associated with the fitting criterion,
- a selection criterion:

  - 'best': select the distribution minimizing (or maximizing, depending
    on the criterion) the criterion,
  - 'first': Select the first distribution for which the criterion is
    greater (or lower, depending on the criterion) than the level,

- a name for the :class:`.ParametricStatistics` object (by default,
  the name is the concatenation of 'ParametricStatistics' and
  and the name of the :class:`.Dataset`).

Capabilities
------------

By inheritance, a :class:`.ParametricStatistics` object has the
same capabilities as :class:`.Statistics`. Additional ones are:

- :meth:`.get_fitting_matrix`: this method shows the values
  of the fitting criterion for the different variables and
  candidate probability distributions
  as well as the select probability distribution,
- :meth:`.plot_criteria`: this method plots the criterion values
  for a given variable.
"""
from __future__ import absolute_import, division, unicode_literals

import os

import matplotlib.pyplot as plt
import openturns as ot
from future import standard_library
from numpy import array, exp, inf, linspace, log
from past.utils import old_div

from gemseo.third_party.prettytable.prettytable import PrettyTable
from gemseo.uncertainty.distributions.ot_dist import OTNormalDistribution
from gemseo.uncertainty.distributions.ot_fdist import OTDistributionFitter
from gemseo.uncertainty.statistics.statistics import Statistics

standard_library.install_aliases()

from gemseo import LOGGER


class ParametricStatistics(Statistics):
    """ Parametric estimation of statistics. """

    def __init__(
        self,
        dataset,
        distributions,
        variables_names=None,
        fitting_criterion="BIC",
        level=0.05,
        selection_criterion="best",
        name=None,
    ):
        """Constructor

        :param Dataset dataset: dataset
        :param list(str) distributions: list of distributions names
        :param list(str) variables_names: list of variables names
            or list of variables names. If None, the method considers
            all variables from loaded dataset. Default: None.
        :param str fitting_criterion: goodness-of-fit criterion.
            Default: 'BIC'.
        :param float level: risk of committing a Type 1 error,
            that is an incorrect rejection of a true null hypothesis,
            for criteria based on test hypothesis.
            Default: 0.05.
        :param str selection_criterion: selection criterion. Default: 'best'
        :param str name: name of the object.
            If None, use the concatenation of class and dataset names.
            Default: None.
        """
        super(ParametricStatistics, self).__init__(dataset, variables_names, name)
        significance_tests = OTDistributionFitter.SIGNIFICANCE_TESTS
        self.fitting_criterion = fitting_criterion
        self.selection_criterion = selection_criterion
        LOGGER.info("| Set goodness-of-fit criterion: %s.", fitting_criterion)
        if self.fitting_criterion in significance_tests:
            self.level = level
            LOGGER.info("| Set significance level of hypothesis test: %s.", level)
        else:
            self.level = None
        self.build_distributions(distributions)

    def build_distributions(self, distributions):
        """Build distributions from a list of distributions names,
        a test level and the stored dataset.

        :param list(str) distributions: list of distributions names.
        """
        self._all_distributions = self._fit_distributions(distributions)
        self.distributions = self._select_best_distributions(distributions)

    def get_fitting_matrix(self):
        """Get the fitting matrix. This matrix contains goodness-of-fit
        measures for each pair < variable, distribution >."""
        rownames = sorted(self._all_distributions.keys())
        colnames = list(self._all_distributions[rownames[0]].keys())
        table = PrettyTable(["Variable"] + colnames + ["Selection"])
        for varname in rownames:
            row, _ = self.get_criteria(varname)
            row = [varname] + [row[distribution] for distribution in colnames]
            row = row + [self.distributions[varname]["name"]]
            table.add_row(row)
        return str(table)

    def get_criteria(self, varname):
        """Get criteria for a given variable name.

        :param str varname: variable name.
        """
        varname_dist = self._all_distributions[varname]
        criteria = {
            distribution: result["criterion"]
            for distribution, result in varname_dist.items()
        }
        is_pvalue = False
        significance_tests = OTDistributionFitter.SIGNIFICANCE_TESTS
        if self.fitting_criterion in significance_tests:
            criteria = {
                distribution: result[1]["p-value"]
                for distribution, result in criteria.items()
            }
            is_pvalue = True
        return criteria, is_pvalue

    def plot_criteria(
        self, varname, title=None, save=False, show=True, n_legend_cols=4, directory="."
    ):
        """Plot criteria for a given variable name

        :param str varname: name of the variable
        :param str title: title. Default: None.
        :param bool save: save the plot into a file. Default: False.
        :param bool show: show the plot. Default: True.
        :param int n_legend_cols: number of text columns in the upper legend.
            Default: 4.
        :param str directory: directory absolute or relative path.
            Default: '.'.
        """
        if varname not in self.names:
            raise ValueError(
                varname + " is not a variable of the dataset."
                "Available ones are:" + ", ".join(self.names)
            )
        criteria, is_pvalue = self.get_criteria(varname)
        xvals = []
        yvals = []
        labels = []
        xval = 0
        for distribution, criterion in criteria.items():
            xval += 1
            xvals.append(xval)
            yvals.append(criterion)
            labels.append(distribution)
        plt.subplot(121)
        plt.bar(xvals, yvals, tick_label=labels, align="center")
        if is_pvalue:
            plt.ylabel("p-value from " + self.fitting_criterion + " test")
            plt.axhline(self.level, color="r", linewidth=2.0)
        plt.grid(True, "both")
        plt.subplot(122)
        data = array(self.dataset[varname])
        data_min = min(data)
        data_max = max(data)
        xvals = linspace(data_min, data_max, 1000)
        distributions = self._all_distributions[varname]
        try:
            plt.hist(data, density=True)
        except AttributeError:
            plt.hist(data, normed=True)
        for dist_name, dist_value in distributions.items():
            pdf = dist_value["fitted_distribution"].distribution.computePDF
            yvals = [pdf([xval])[0] for xval in xvals]
            plt.plot(xvals, yvals, label=dist_name, linewidth=2.0)
        plt.legend(
            bbox_to_anchor=(0.0, 1.02, 1.0, 0.102),
            loc="lower left",
            ncol=n_legend_cols,
            mode="expand",
            borderaxespad=0.0,
        )
        plt.grid(True, "both")
        if title is not None:
            plt.suptitle(title)
        filename = os.path.join(directory, "criteria.pdf")
        if save:
            plt.savefig(filename)
        if show:
            plt.show()
        plt.close()

    def _select_best_distributions(self, distributions_names):
        """ Select the best distributions for the different variables."""
        LOGGER.info("Select the best distribution for each variable.")
        distributions = {}
        for varname in self.names:
            varname_dist = self._all_distributions[varname]
            criteria = [
                varname_dist[distribution]["criterion"]
                for distribution in distributions_names
            ]
            select_from_results = OTDistributionFitter.select_from_results
            index = select_from_results(
                criteria, self.fitting_criterion, self.level, self.selection_criterion
            )
            name = distributions_names[index]
            value = varname_dist[name]["fitted_distribution"]
            distributions[varname] = {"name": name, "value": value}
            LOGGER.info("| The best distribution for %s is %s.", varname, value)
        return distributions

    def _fit_distributions(self, distributions):
        """Fit distributions for the different dataset marginals
        among a given collection of distributions names.

        :param list(str) distributions: list of distributions names
        """
        dist_list = ", ".join(distributions)
        LOGGER.info(
            "Fit different distributions (%s) per variable "
            "and compute the goodness-of-fit criterion.",
            dist_list,
        )
        results = {}
        for varname in self.names:
            LOGGER.info("| Fit different distributions for %s.", varname)
            dataset = self.dataset[varname]
            results[varname] = self._fit_marginal_distributions(
                varname, dataset, distributions
            )
        return results

    def _fit_marginal_distributions(self, variable, sample, distributions):
        """Fit distributions for a given dataset marginal
        among a given collection of distributions names

        :param str variable: variable name
        :param array dataset: sample
        :param list(str) distributions: list of distributions names
        """
        result = {}
        factory = OTDistributionFitter(variable, sample)
        for distribution in distributions:
            fitted_distribution = factory.fit(distribution)
            test_result = factory.measure(
                fitted_distribution, self.fitting_criterion, self.level
            )
            result[distribution] = {}
            result[distribution]["fitted_distribution"] = fitted_distribution
            result[distribution]["criterion"] = test_result
        return result

    def maximum(self):
        """Get the maximum.

        :return: maximum
        """
        result = {
            name: self.distributions[name]["value"].math_upper_bound
            for name in self.names
        }
        return result

    def mean(self):
        """Get the mean.

        :return: mean
        """
        result = {name: self.distributions[name]["value"].mean for name in self.names}
        return result

    def minimum(self):
        """Get the minimum.

        :return: minimum
        """
        result = {
            name: self.distributions[name]["value"].math_lower_bound
            for name in self.names
        }
        return result

    def probability(self, thresh, greater=False):
        """Compute a probability associated to a threshold.

        :param float thresh: threshold
        :param bool greater: if True, compute the probability the probability
            of exceeding the threshold, if False, compute the reverse.
            Default: True.
        :return: probability
        """
        dist = self.distributions
        if greater:
            result = {
                name: 1 - dist[name]["value"].cdf(thresh[name])[0]
                for name in self.names
            }
        else:
            result = {
                name: dist[name]["value"].cdf(thresh[name])[0] for name in self.names
            }
        return result

    def tolerance_interval(self, coverage, confidence=0.95, side="both"):
        """Compute the tolerance interval (TI) for a given minimum percentage
        of the population and a given confidence level.

        :param float coverage: minimum percentage of belonging to the TI.
        :param float confidence: level of confidence in [0,1]. Default: 0.95.
        :param str side: kind of interval: 'lower' for lower-sided TI,
            'upper' for upper-sided TI and 'both for both-sided TI.
        :return: tolerance limits
        """
        if side not in ["upper", "lower", "both"]:
            raise ValueError(
                "The argument 'side' represents the type"
                "of tolerance bounds. Available ones are: 'upper'"
                "'lower' and 'both'."
            )
        if not 0.0 <= coverage <= 1.0:
            raise ValueError("The argument 'coverage'" " must be number in [0,1].")
        if not 0.0 <= confidence <= 1.0:
            raise ValueError("The argument 'confidence'" " must be number in [0,1].")
        limits = {}
        for varname in self.names:
            dist_name = self.distributions[varname]["name"]
            dist = self.distributions[varname]["value"]
            if dist_name == "Normal":
                lower, upper = self._normal_tolerance_interval(
                    dist, coverage, confidence, side
                )
            elif dist_name == "Uniform":
                lower, upper = self._uniform_tolerance_interval(
                    dist, coverage, confidence, side
                )
            elif dist_name == "LogNormal":
                lower, upper = self._lognormal_tolerance_interval(
                    dist, coverage, confidence, side
                )
            elif dist_name == "WeibullMin":
                lower, upper = self._weibull_tolerance_interval(
                    dist, coverage, confidence, side
                )
            elif dist_name == "Exponential":
                lower, upper = self._exponential_tolerance_interval(
                    dist, coverage, confidence, side
                )
            else:
                raise ValueError(
                    "Tolerance interval is not implemented for"
                    'distribution "' + dist_name + '".'
                )
            limits[varname] = (lower, upper)
        return limits

    def _weibull_tolerance_interval(
        self, dist, coverage, confidence=0.95, side="lower"
    ):
        """Compute the tolerance interval (TI) for a given minimum percentage
        of the population and a given confidence level, when data are
        Weibull distributed

        :param WeibullMin dist: OT WeibullMin distribution
        :param float coverage: minimum percentage of belonging to the TI.
        :param float confidence: level of confidence in [0,1]. Default: 0.95.
        :param str side: kind of interval: 'lower' for one-sided lower TI,
            'upper' for one-sided upper TI or 'both' for two-sided TI.
        :return: tolerance limits
        """
        alpha = dist.marginals[0].getParameter()[0]
        beta = dist.marginals[0].getParameter()[1]
        gamma = dist.marginals[0].getParameter()[2]
        x_i = log(beta)
        delta = 1.0 / alpha
        if side == "upper":
            lbd = log(-log(1 - coverage))
            offset = -self.n_samples ** 0.5 * lbd
            dof = self.n_samples - 1
            student = ot.Student(dof, offset)
            upper = delta * student.computeQuantile(confidence)[0]
            upper /= (self.n_samples - 1) ** 0.5
            upper = x_i - upper
            lower = -inf
        elif side == "lower":
            lbd = log(-log(coverage))
            offset = -self.n_samples ** 0.5 * lbd
            dof = self.n_samples - 1
            student = ot.Student(dof, offset)
            lower = delta * student.computeQuantile(1 - confidence)[0]
            lower /= (self.n_samples - 1) ** 0.5
            lower = x_i - lower
            upper = inf
        else:
            coverage = (coverage + 1) / 2.0
            alpha = (1 - confidence) / 2.0
            lbd = log(-log(1 - coverage))
            offset = -self.n_samples ** 0.5 * lbd
            dof = self.n_samples - 1
            student = ot.Student(dof, offset)
            upper = delta * student.computeQuantile(1 - alpha)[0]
            upper /= (self.n_samples - 1) ** 0.5
            upper = x_i - upper
            lbd = log(-log(coverage))
            offset = -self.n_samples ** 0.5 * lbd
            dof = self.n_samples - 1
            student = ot.Student(dof, offset)
            lower = delta * student.computeQuantile(alpha)[0]
            lower /= (self.n_samples - 1) ** 0.5
            lower = x_i - lower
        limits = (array([exp(lower) + gamma]), array([exp(upper) + gamma]))
        return limits

    def _uniform_tolerance_interval(
        self, dist, coverage, confidence=0.95, side="lower"
    ):
        """Compute the tolerance interval (TI) for a given minimum percentage
        of the population and a given confidence level, when data are
        uniformly distributed

        :param Distribution dist: OT uniform distribution
        :param float coverage: minimum percentage of belonging to the TI.
        :param float confidence: level of confidence in [0,1]. Default: 0.95.
        :param str side: kind of interval: 'lower' for one-sided lower TI,
            'upper' for one-sided upper TI or 'both' for two-sided TI.
        :return: tolerance limits
        """
        minimum = dist.marginals[0].getParameter()[0]
        maximum = dist.marginals[0].getParameter()[1]
        if side == "upper":
            upper = (maximum - minimum) * coverage
            upper /= (1 - confidence) ** (1.0 / self.n_samples)
            upper += minimum
            limits = (array([-inf]), array([upper]))
        elif side == "lower":
            lower = (maximum - minimum) * (1 - coverage)
            lower /= confidence ** (1.0 / self.n_samples)
            lower += minimum
            limits = (array([lower]), array([inf]))
        else:
            upper = (maximum - minimum) * (coverage + 1) / 2.0
            upper /= ((1 - confidence) / 2.0) ** (1.0 / self.n_samples)
            upper += minimum
            lower = (maximum - minimum) * (1 - (coverage + 1) / 2.0)
            lower /= (1 - (1 - confidence) / 2.0) ** (1.0 / self.n_samples)
            lower += minimum
            limits = (array([lower]), array([upper]))
        return limits

    def _exponential_tolerance_interval(
        self, dist, coverage, confidence=0.95, side="lower"
    ):
        """Compute the tolerance interval (TI) for a given minimum percentage
        of the population and a given confidence level, when data are
        exponentially distributed

        :param Exponential dist: OT exponential distribution
        :param float coverage: minimum percentage of belonging to the TI.
        :param float confidence: level of confidence in [0,1]. Default: 0.95.
        :param str side: kind of interval: 'lower' for one-sided lower TI,
            'upper' for one-sided upper TI or 'both' for two-sided TI.
        :return: tolerance limits
        """
        lmbda = dist.marginals[0].getParameter()[0]
        gamma = dist.marginals[0].getParameter()[1]
        if side == "upper":
            chisq = ot.ChiSquare(2 * self.n_samples)
            chisq = chisq.computeQuantile(confidence)[0]
            upper = -2 * self.n_samples * log(coverage) * lmbda
            upper /= chisq
            lower = 0.0
        elif side == "lower":
            chisq = ot.ChiSquare(2 * self.n_samples)
            chisq = chisq.computeQuantile(confidence)[0]
            lower = -2 * self.n_samples * log(1 - coverage) * lmbda
            lower /= chisq
            upper = inf
        else:
            coverage = (coverage + 1) / 2.0
            alpha = (1 - confidence) / 2.0
            chisq = ot.ChiSquare(2 * self.n_samples)
            chisq = chisq.computeQuantile(1 - alpha)[0]
            upper = -2 * self.n_samples * log(1 - coverage) * lmbda
            upper /= chisq
            lower = -2 * self.n_samples * log(coverage) * lmbda
            lower /= chisq

        limits = (array([lower + gamma]), array([upper + gamma]))
        return limits

    def _normal_tolerance_interval(self, dist, coverage, confidence=0.95, side="lower"):
        """Compute the tolerance interval (TI) for a given minimum percentage
        of the population and a given confidence level, when data are
        normally distributed

        :param Normal dist: OT normal distribution
        :param float coverage: minimum percentage of belonging to the TI.
        :param float confidence: level of confidence in [0,1]. Default: 0.95.
        :param str side: kind of interval: 'lower' for one-sided lower TI,
            'upper' for one-sided upper TI or 'both' for two-sided TI.
        :return: tolerance limits
        """
        mean = dist.marginals[0].getParameter()[0]
        std = dist.marginals[0].getParameter()[1]
        if side in ["upper", "lower"]:
            z_p = ot.Normal().computeQuantile(coverage)[0]
            delta = z_p * self.n_samples ** 0.5
            dof = self.n_samples - 1
            student = ot.Student(dof, delta)
            student_quantile = student.computeQuantile(confidence)[0]
            tolerance_factor = old_div(student_quantile, self.n_samples ** 0.5)
            if side == "upper":
                upper = mean + tolerance_factor * std
                limits = (array([-inf]), array([upper]))
            else:
                lower = mean - tolerance_factor * std
                limits = (array([lower]), array([inf]))
        else:
            z_p = ot.Normal().computeQuantile((1 + coverage) / 2.0)[0]
            left = (1 + 1.0 / self.n_samples) ** 0.5 * z_p
            chisq = ot.ChiSquare(self.n_samples - 1)
            right = old_div(
                (self.n_samples - 1), chisq.computeQuantile(1 - confidence)[0]
            )
            right = right ** 0.5
            weight = self.n_samples - 3 - chisq.computeQuantile(1 - confidence)[0]
            weight /= 2 * (self.n_samples + 1) ** 2
            weight = (1 + weight) ** 0.5
            tolerance_factor = left * right * weight
            lower = mean - tolerance_factor * std
            upper = mean + tolerance_factor * std
            limits = (array([lower]), array([upper]))
        return limits

    def _lognormal_tolerance_interval(
        self, dist, coverage, confidence=0.95, side="lower"
    ):
        """Compute the tolerance interval (TI) for a given minimum percentage
        of the population and a given confidence level, when data are
        normally distributed

        :param LogNormal dist: OT LogNormal distribution
        :param float coverage: minimum percentage of belonging to the TI.
        :param float confidence: level of confidence in [0,1]. Default: 0.95.
        :param str side: kind of interval: 'lower' for one-sided lower TI,
            'upper' for one-sided upper TI or 'both' for two-sided TI.
        :return: tolerance limits
        """
        dist = OTNormalDistribution(
            "x",
            dist.marginals[0].getParameter()[0],
            dist.marginals[0].getParameter()[1],
        )
        lower, upper = self._normal_tolerance_interval(dist, coverage, confidence, side)
        limits = (exp(lower), exp(upper))
        return limits

    def quantile(self, prob):
        """Get the quantile associated to a given probability.

        :param float prob: probability
        :return: quantile
        :rtype: float or list(float)
        """
        prob = array([prob])
        result = {
            name: self.distributions[name]["value"].inverse_cdf(prob)
            for name in self.names
        }
        return result

    def standard_deviation(self):
        """Get the standard deviation.

        :return: standard deviation
        :rtype: float or list(float)
        """
        result = {
            name: self.distributions[name]["value"].standard_deviation
            for name in self.names
        }
        return result

    def variance(self):
        """Get the variance.

        :return: variance
        :rtype: float or list(float)
        """
        result = {
            name: self.distributions[name]["value"].standard_deviation ** 2
            for name in self.names
        }
        return result

    def moment(self, order):
        """Compute the moment for a given order, either centered or not.

        :param int order: moment index
        :return: moment
        :rtype: float or list(float)
        """
        dist = self.distributions
        result = [
            dist[varname]["value"].distribution.getMoment(order)[0]
            for varname in self.names
        ]
        return result

    def range(self):
        """Get the range of variables.

        :return: range of variables
        """
        result = {}
        for name in self.names:
            dist = self.distributions[name]["value"]
            result[name] = dist.math_upper_bound - dist.math_lower_bound
        return result

    @classmethod
    def get_available_distributions(cls):
        """ Get available distributions. """
        return sorted(OTDistributionFitter.AVAILABLE_FACTORIES.keys())

    @classmethod
    def get_available_criteria(cls):
        """ Get available goodness-of-fit criteria. """
        return sorted(OTDistributionFitter.AVAILABLE_FITTING_TESTS.keys())

    @classmethod
    def get_significance_tests(cls):
        """ Get significance tests. """
        return sorted(OTDistributionFitter.SIGNIFICANCE_TESTS)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalable design space from Tedford and Martins (2010)
*****************************************************

"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import zeros

from gemseo.algos.design_space import DesignSpace
from gemseo.problems.scalable.scalable_tm.variables import (
    X_SHARED_NAME,
    check_consistency,
    get_coupling_name,
    get_x_local_name,
)

standard_library.install_aliases()

from gemseo import LOGGER


class TMDesignSpace(DesignSpace):

    """The design space for the scalable problem introduced by Tedford and
    Martins (2010) defines the lower and upper bounds of both local design
    parameters, shared design parameters and coupling variables, as well as
    default values. The lower bounds are all equal to 0, the upper bounds are
    all equal to 1 and default values are all equal to 0.5.
    """

    def __init__(
        self,
        n_shared=1,
        n_local=None,
        n_coupling=None,
        default_inputs=None,
        dtype="float64",
    ):
        """The TMDesignSpace constructor requires the number of shared design
        parameters, the number of local design parameters per discipline
        and the number of coupling variables per discipline. The two latter
        arguments must be list of integers with the same length which
        corresponds to the number of strongly coupled disciplines. By default,
        the design space considers two disciplines.

        :param int n_shared: size of the shared design parameters.
            Default: 1.
        :param list(int) n_local: sizes of the local design parameters.
            If None, use [1, 1]. Default: None.
        :param list(int) n_coupling: sizes of the coupling parameters.
            If None, use [1, 1]. Default: None.
        :param dict default_inputs: default inputs.
            Default: None.
        :param dtype: numpy data type. Default: 'float64'.
        """
        super(TMDesignSpace, self).__init__()
        n_local = n_local or [1, 1]
        n_coupling = n_coupling or [1, 1]
        check_consistency(n_shared, n_local, n_coupling)

        if not isinstance(default_inputs, dict):
            default_inputs = {}
        for id_x, n_x in enumerate(n_local):
            value = default_inputs.get(get_x_local_name(id_x), zeros(n_x) + 0.5)
            value.astype(dtype)
            self.add_variable(
                get_x_local_name(id_x), n_x, l_b=0.0, u_b=1.0, value=value
            )
        value = default_inputs.get(X_SHARED_NAME, zeros(n_shared) + 0.5)
        value.astype(dtype)
        self.add_variable(X_SHARED_NAME, n_shared, l_b=0.0, u_b=1.0, value=value)
        for id_y, n_y in enumerate(n_coupling):
            value = default_inputs.get(get_coupling_name(id_y), zeros(n_y) + 0.5)
            value.astype(dtype)
            self.add_variable(
                get_coupling_name(id_y), n_y, l_b=0.0, u_b=1.0, value=value
            )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                        documentation
#        :author: Francois Gallard, Charlie Vanaret
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Functions, f(x), from disciplines execution
*******************************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

from multiprocessing import Value
from numbers import Number

from future import standard_library
from future.utils import with_metaclass
from numpy import abs as np_abs
from numpy import (
    add,
    atleast_1d,
    atleast_2d,
    hstack,
    inner,
    ndarray,
    reshape,
    subtract,
    vstack,
    where,
)
from numpy.linalg import norm
from six import string_types

from gemseo.utils.data_conversion import DataConversion
from gemseo.utils.derivatives_approx import ComplexStep, FirstOrderFD
from gemseo.utils.singleton import SingleInstancePerAttributeId

standard_library.install_aliases()


from gemseo import LOGGER


class MDOFunction(object):
    """A container object to represent a function in an MDO process"""

    TYPE_OBJ = "obj"
    TYPE_EQ = "eq"
    TYPE_INEQ = "ineq"
    TYPE_OBS = "obs"
    AVAILABLE_TYPES = [TYPE_OBJ, TYPE_EQ, TYPE_INEQ, TYPE_OBS]
    DICT_REPR_ATTR = ["name", "f_type", "expr", "args", "dim"]

    def __init__(
        self,
        func,
        name,
        f_type=None,
        jac=None,
        expr=None,
        args=None,
        dim=None,
        outvars=None,
        force_real=False,
    ):
        """
        Initializes the function attributes

        :param func: the pointer to the function to be actually called
        :param name: the name of the function as a string
        :param f_type: the type of function among (obj, eq, ineq, obs)
        :param jac: the jacobian
        :param expr: the function expression
        :param args: the function arguments
        :param dim: the output dimension
        :param outvars: the array of variable names used as inputs
            of the function
        :param force_real: if True, cast the results to real value
        """
        super(MDOFunction, self).__init__()
        self._n_calls = Value("i", 0)
        # Initialize attributes
        self._f_type = None
        self._func = NotImplementedCallable()
        self._jac = NotImplementedCallable()
        self._name = None
        self._args = None
        self._expr = None
        self._dim = None
        self._outvars = None
        # Use setters to check values
        self.func = func
        self.jac = jac
        self.name = name
        self.f_type = f_type
        self.expr = expr
        self.args = args
        self.dim = dim
        self.outvars = outvars
        self.last_eval = None
        self.force_real = force_real

    @property
    def n_calls(self):
        """
        Returns the number of calls to execute() which triggered the _run()
        multiprocessing safe
        """
        return self._n_calls.value

    @n_calls.setter
    def n_calls(self, value):
        """
        Sets the number of calls to execute() which triggered the _run()
        multiprocessing safe
        """
        with self._n_calls.get_lock():
            self._n_calls.value = value

    @property
    def func(self):
        """Accessor to the func property"""
        return self.__counted_f

    def __counted_f(self, x_vect):
        """
        Calls the function in self[self.SCIPY_FUN_TAG]

        :param x_vect: the function argument
        """
        with self._n_calls.get_lock():
            self._n_calls.value += 1
        val = self._func(x_vect)
        self.last_eval = val
        return val

    @func.setter
    def func(self, f_pointer):
        """Sets the f pointer

        :param f_pointer: the pointer to the function to be actually called
        """
        self._n_calls.value = 0
        self._func = f_pointer

    def __call__(self, x_vect):
        """
        Calls the function

        :param x_vect: the function argument
        """
        val = self.evaluate(x_vect, self.force_real)
        return val

    def evaluate(self, x_vect, force_real=False):
        """
        Evaluate the function

        :param x_vect: the function argument
        :param force_real: if True, cast the results to real value
        """
        val = self.__counted_f(x_vect)
        if force_real:
            val = val.real
        self.dim = atleast_1d(val).shape[0]
        return val

    @property
    def name(self):
        """Accessor to the name of the function"""
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of the function

        :param name: the name of the functioon
        """
        if name is not None and not isinstance(name, string_types):
            raise TypeError(
                "MDOFunction name must be a string, got " + str(type(name)) + " instead"
            )
        self._name = name

    @property
    def f_type(self):
        """Accessor to the function type"""
        return self._f_type

    @f_type.setter
    def f_type(self, f_type):
        """Sets the function type

        :param f_type: the type of function among MDOFunction.AVAILABLE_TYPES
        """
        if f_type is not None and f_type not in self.AVAILABLE_TYPES:
            raise ValueError(
                "MDOFunction type must be among "
                + str(self.AVAILABLE_TYPES)
                + " , got "
                + str(f_type)
                + " instead"
            )
        self._f_type = f_type

    @property
    def jac(self):
        """Accessor to the jacobian"""
        return self._jac

    @jac.setter
    def jac(self, jac):
        """Sets the function jacobian

        :param jac: pointer to a jacobian function
        """
        if jac is not None:
            if not callable(jac):
                raise TypeError("Jacobian function must be callable")
        self._jac = jac

    @property
    def args(self):
        """Accessor to the arguments list"""
        return self._args

    @args.setter
    def args(self, args):
        """Sets the function arguments

        :param args: arguments list
        """
        if args is not None:
            if isinstance(args, ndarray):
                self._args = args.tolist()
            else:
                self._args = list(args)
        else:
            self._args = None

    @property
    def expr(self):
        """Accessor to the expression"""
        return self._expr

    @expr.setter
    def expr(self, expr):
        """Sets the function expression

        :param expr: arguments list
        """
        if expr is not None and not isinstance(expr, string_types):
            raise TypeError(
                "Expression must be a string. "
                + " Got "
                + str(type(expr))
                + " instead."
            )
        self._expr = expr

    @property
    def dim(self):
        """Accessor to the dimension"""
        return self._dim

    @dim.setter
    def dim(self, dim):
        """Sets the function dimension

        :param dim: int
        """
        if dim is not None and not int(dim) == dim:
            raise TypeError(
                "Dimension must be a integer. " + " Got " + str(type(int)) + " instead."
            )
        self._dim = dim

    @property
    def outvars(self):
        """Accessor to the array of input variable names
        used by the function
        """
        return self._outvars

    @outvars.setter
    def outvars(self, outvars):
        """Sets the array of output variable names used by the function

        :param outvars: array of variable names
        """
        if outvars is not None:
            if isinstance(outvars, ndarray):
                self._outvars = outvars.tolist()
            else:
                self._outvars = list(outvars)
        else:
            self._outvars = None

    def is_constraint(self):
        """Returns True if self.f_type is eq or ineq"""
        return self.f_type in [self.TYPE_EQ, self.TYPE_INEQ]

    def __repr__(self):
        """
        Self representation as a string
        """
        str_repr = self.name
        if self.has_args():
            str_repr += "(" + (", ".join(self.args)) + ")"

        if self.has_expr():
            str_repr += " = "
            expr = self.expr
            n_char = len(str_repr)
            # Remove empty lines with filter
            expr_spl = [_f for _f in expr.split("\n") if _f]
            str_repr += expr_spl[0]
            for repre in expr_spl[1:]:
                str_repr += "\n" + " " * n_char + repre

        return str_repr

    def has_jac(self):
        """Check if MDOFunction has a jacobian

        :returns: True if self has a jacobian
        """
        return self.jac is not None and not isinstance(
            self._jac, NotImplementedCallable
        )

    def has_dim(self):
        """Check if MDOFunction has a dimension

        :returns: True if self has a dimension
        """
        return self.dim is not None

    def has_outvars(self):
        """Check if MDOFunction has an array of input variables

        :returns: True if self has a dimension
        """
        return self.outvars is not None

    def has_expr(self):
        """Check if MDOFunction has an expression

        :returns: True if self has an expression
        """
        return self.expr is not None

    def has_args(self):
        """Check if MDOFunction has an args

        :returns: True if self has an args
        """
        return self.args is not None

    def has_f_type(self):
        """Check if MDOFunction has an type

        :returns: True if self has a type
        """
        return self.f_type is not None

    def apply_operator(self, other_f, operator, operator_repr):
        """Defines addition/substraction  for MDOFunction
        Supports automatic differentiation if other_f and self have a Jacobian

        :param other_f: param operator:
        :param operator_repr: the representation as a string
        :param operator: the operator as a function pointer
        """
        if isinstance(other_f, MDOFunction):
            return self.__apply_operator_tofunction(other_f, operator, operator_repr)

        if isinstance(other_f, Number):
            return self.__apply_operator_to_number(other_f, operator, operator_repr)

        raise TypeError(
            "Unuspported + operand for MDOFunction and " + str(type(other_f))
        )

    def __apply_operator_tofunction(self, other_f, operator, operator_repr):
        """Defines addition/substraction  for MDOFunction
        Supports automatic differentiation if other_f and self have a Jacobian

        :param other_f: param operator:
        :param operator_repr: the representation as a string
        :param operator: the operator as a function pointer
        """
        assert isinstance(other_f, MDOFunction)

        def add_f_pt(x_vect):
            """Addition function

            :param x_vect: design variables
            :returns: self(x_vect) added to other_f(x_vect)
            """
            selfval = self(x_vect)
            otherval = other_f(x_vect)
            return operator(selfval, otherval)

        add_name = self.name + operator_repr + other_f.name
        self_oper_f = MDOFunction(
            add_f_pt,
            add_name,
            args=self.args,
            f_type=self.f_type,
            dim=self.dim,
            outvars=self.outvars,
        )

        if self.has_jac() and other_f.has_jac():

            def add_jac(x_vect):
                """Jacobian of addition function

                :param x_vect: design variables
                :returns: Jac self(x_vect) added to Jac other_f(x_vect)
                """
                self_jac = self._jac(x_vect)
                other_jac = other_f.jac(x_vect)
                return operator(self_jac, other_jac)

            self_oper_f.jac = add_jac

        if self.has_expr() and other_f.has_expr():
            f_expr = self.expr + operator_repr + other_f.expr
            self_oper_f.expr = f_expr

        if self.has_args() and other_f.has_args():
            args = sorted(list(set(self.args + other_f.args)))
            self_oper_f.args = args

        if self.has_f_type():
            self_oper_f.f_type = self.f_type
        elif other_f.has_f_type():
            self_oper_f.f_type = other_f.f_type

        return self_oper_f

    def __apply_operator_to_number(self, other_val, operator, operator_repr):
        """
        Defines addition/substraction  for MDOFunction
        Supports automatic differentiation if other_f and self have a Jacobian

        :param other_val: param operator: the value to add/subs
        :param operator_repr: the representation as a string
        :param operator: the operator as a function pointer
        """
        assert isinstance(other_val, Number)

        def add_f_pt(x_vect):
            """Addition function

            :param x_vect: design variables
            :returns: self(x_vect) added to other_f(x_vect)
            """
            selfval = self(x_vect)
            return operator(selfval, other_val)

        add_name = self.name + operator_repr + str(other_val)
        self_oper_f = MDOFunction(
            add_f_pt,
            add_name,
            args=self.args,
            f_type=self.f_type,
            dim=self.dim,
            outvars=self.outvars,
        )

        if self.has_jac():
            self_oper_f.jac = self.jac

        if self.has_expr():
            self_oper_f.expr = self.expr + operator_repr + str(other_val)

        if self.has_args():
            self_oper_f.args = self.args

        if self.has_f_type():
            self_oper_f.f_type = self.f_type

        return self_oper_f

    def __add__(self, other_f):
        """
        Defines addition for MDOFunction
        Supports automatic differentiation if other_f and self have a jacobian
        """
        return self.apply_operator(other_f, operator=add, operator_repr="+")

    def __sub__(self, other_f):
        """
        Defines subtraction for MDOFunction
        Supports automatic differentiation if other_f and self have a jacobian
        """
        return self.apply_operator(other_f, operator=subtract, operator_repr="-")

    def __neg__(self):
        """
        Defines minus operator for MDOFunction (-f)
        Supports automatic differentiation if other_f and self have a Jacobian
        """

        def min_pt(x_vect):
            """Negative self function

            :param x_vect: design variables
            :returns: self(x_vect)
            """
            return -self(x_vect)

        min_name = "-" + self.name
        min_self = MDOFunction(
            min_pt,
            min_name,
            args=self.args,
            f_type=self.f_type,
            dim=self.dim,
            outvars=self.outvars,
        )

        if self.has_jac():

            def min_jac(x_vect):
                """Jacobian of negative self

                :param x_vect: design variables
                :returns: Jac of -self(x_vect)
                """
                return -self.jac(x_vect)  # pylint: disable=E1102

            min_self.jac = min_jac

        if self.has_expr():
            min_self.expr = "-" + self.expr

        return min_self

    def __mul__(self, other):
        """
        Defines multiplication for MDOFunction
        Supports automatic linearization if other_f and self have a jacobian
        """
        is_number = isinstance(other, Number)
        is_func = isinstance(other, MDOFunction)
        if not is_number and not is_func:
            raise TypeError(
                "Unuspported * operand for MDOFunction and " + str(type(other))
            )

        if is_func:

            def mul_f_pt(x_vect):
                """Multiplication function

                :param x_vect: design variables
                :returns: self(x_vect) added to other_f(x_vect)
                """
                return self(x_vect) * other(x_vect)

            out_name = self.name + "*" + other.name

        else:  # Func is a number

            def mul_f_pt(x_vect):
                """Multiplication function

                :param x_vect: design variables
                :returns: self(x_vect) added to other_f(x_vect)
                """
                return self(x_vect) * other

            out_name = self.name + "*" + str(other)

        self_oper_f = MDOFunction(
            mul_f_pt,
            out_name,
            args=self.args,
            f_type=self.f_type,
            dim=self.dim,
            outvars=self.outvars,
        )
        if is_func:
            if self.has_jac() and other.has_jac():

                def mult_jac(x_vect):
                    """Jacobian of multiplication function

                    :param x_vect: design variables
                    :returns: Jac self(x_vect) added to Jac other_f(x_vect)
                    """
                    self_f = self(x_vect)
                    other_f = other(x_vect)
                    self_jac = self._jac(x_vect)
                    other_jac = other.jac(x_vect)
                    return self_jac * other_f + other_jac * self_f

                self_oper_f.jac = mult_jac
        elif self.has_jac():

            def mult_jac(x_vect):
                """Jacobian of multiplication function

                :param x_vect: design variables
                :returns: Jac self(x_vect) added to Jac other_f(x_vect)
                """
                return self._jac(x_vect) * other

            self_oper_f.jac = mult_jac

        if self.has_expr():
            if is_func and other.has_expr():
                f_expr = "(" + self.expr + ")*(" + other.expr + ")"
                self_oper_f.expr = f_expr
            if is_number:
                f_expr = "(" + self.expr + ")*" + str(other)
                self_oper_f.expr = f_expr

        if is_func and self.has_args() and other.has_args():
            args = sorted(list(set(self.args + other.args)))
            self_oper_f.args = args

        if self.has_f_type():
            self_oper_f.f_type = self.f_type
        elif is_func and other.has_f_type():
            self_oper_f.f_type = other.f_type

        return self_oper_f

    def offset(self, value):
        """Adds an offset value to the function

        :param value: the offset value
        :returns: the offset function as an MDOFunction
        """

        def wrapped_function(x_vect):
            """Wrapped provided function in order to give to
            optimizer

            :param x_vect: design variables
            :returns: evaluation of function at design variables
                     plus the offset
            """
            return self(x_vect) + value

        expr = self.expr
        if expr is not None:
            expr += " + " + str(value)
        else:
            expr = self.name + " + " + str(value)
        offset_func = MDOFunction(
            wrapped_function,
            name=self.name,
            f_type=self.f_type,
            expr=expr,
            args=self.args,
            dim=self.dim,
            jac=self.jac,
            outvars=self.outvars,
        )

        return offset_func
        # Nothing to do for Jacobian since it is unchanged.

    def check_grad(self, x_vect, method="FirstOrderFD", step=1e-6, error_max=1e-8):
        """Checks the gradient of self

        :param x_vect: the vector at which the function is checked
        :param method: FirstOrderFD or ComplexStep
            (Default value = "FirstOrderFD")
        :param step: the step for approximation
            (Default value = 1e-6)
        :param error_max: Default value = 1e-8)
        """

        def rel_err(a_vect, b_vect):
            """Compute relative error

            :param a_vect: param b_vect:
            :param b_vect:

            """
            n_b = norm(b_vect)
            if n_b > error_max:
                return norm(a_vect - b_vect) / norm(b_vect)
            return norm(a_vect - b_vect)

        def filt_0(arr, floor_value=1e-6):
            """Filtering of error

            :param arr:
            :param floor_value:  (Default value = 1e-6)
            :param floor_value:  (Default value = 1e-6)
            """
            return where(np_abs(arr) < floor_value, 0.0, arr)

        if method == "FirstOrderFD":
            apprx = FirstOrderFD(self, step)
        elif method == "ComplexStep":
            apprx = ComplexStep(self, step)
        else:
            raise ValueError(
                "Unknwon approximation method "
                + str(method)
                + ", use FirstOrderFD or ComplexStep."
            )
        apprx_grad = apprx.f_gradient(x_vect).real
        anal_grad = self._jac(x_vect).real
        if not apprx_grad.shape == anal_grad.shape:
            shape_app1d = len(apprx_grad.shape) == 1 or apprx_grad.shape[0] == 1
            shape_ex1d = len(anal_grad.shape) == 1 or anal_grad.shape[0] == 1
            shape_1d = shape_app1d and shape_ex1d
            flatten_diff = anal_grad.flatten().shape != apprx_grad.flatten().shape
            if not (shape_1d) or (shape_1d and flatten_diff):
                raise ValueError(
                    "Inconsistent function jacobian shape ! "
                    + "Got :"
                    + str(anal_grad.shape)
                    + " while expected :"
                    + str(apprx_grad.shape)
                )
        succeed = rel_err(anal_grad, apprx_grad) < error_max
        if not succeed:
            LOGGER.error("Function jacobian is wrong %s", str(self))
            LOGGER.error("Error =\n%s", str(filt_0(anal_grad - apprx_grad)))
            LOGGER.error("Analytic jacobian=\n%s", str(filt_0(anal_grad)))
            LOGGER.error("Approximate step gradient=\n%s", str(filt_0(apprx_grad)))
            raise ValueError("Function jacobian is wrong " + str(self))

    def get_data_dict_repr(self):
        """Returns a dict representation of self for serialization
        Pointers to functions are removed

        :returns: a dict with attributes names as keys
        """
        repr_dict = {}
        for attr_name in self.DICT_REPR_ATTR:
            attr = getattr(self, attr_name)
            if attr is not None:
                repr_dict[attr_name] = attr
        return repr_dict

    @staticmethod
    def init_from_dict_repr(**kwargs):
        """Initalizes a new Function from a data dict
        typically used for deserialization

        :param kwargs: key value pairs from DICT_REPR_ATTR
        """
        allowed = MDOFunction.DICT_REPR_ATTR
        for key in kwargs:
            if key not in allowed:
                raise ValueError(
                    "Cannot initialize MDOFunction "
                    + "attribute: "
                    + str(key)
                    + ", allowed ones are: "
                    + ", ".join(allowed)
                )
        return MDOFunction(func=None, **kwargs)

    def set_pt_from_database(
        self, database, design_space, normalize=False, jac=True, x_tolerance=1e-10
    ):
        """
        self.__call__(x) returns f(x) if x is in the database
        and self.name in the database keys
        Idem for jac

        :param database: the database to read
        :param design_space: the design space used for normalization
        :param normalize: if True, x_n is unnormalized before call
        :param jac: if True, a jacobian pointer is also generated
        """
        name = self.name

        def read_in_db(x_n, fname):
            """
            Reads fname in the database for the x_n entry

            :param x_n: the x to evaluate fname
            :returns : f(x_n) if present in the fatabase
            """
            if normalize:
                x_db = design_space.unnormalize_vect(x_n)
            else:
                x_db = x_n
            val = database.get_f_of_x(fname, x_db, x_tolerance)
            if val is None:
                msg = "Function " + str(fname)
                msg += " evaluation relies only on the database, and"
                msg += str(fname) + "( x ) is not in the database "
                msg += " for x=" + str(x_db)
                raise ValueError(msg)
            return val

        def f_from_db(x_n):
            """
            Computes f(x_n) from the database

            :param x_n: the x to evaluate f
            """
            return read_in_db(x_n, name)

        def j_from_db(x_n):
            """
            Computes df(x_n)/d x_n from the database

            :param x_n: the x to evaluate the jacobian
            """
            return read_in_db(x_n, "@" + name)

        self.func = f_from_db
        if jac:
            self.jac = j_from_db


class MDOLinearFunction(MDOFunction):
    """Linear multivariate function with coefficients (a_i)
    f(x) = sum_{i=1}^n a_i * x_i


    """

    def __init__(self, coefficients, name, f_type=None, args=None):
        """
        Initialize the function attributes

        :param coefficients: coefficients of the linear function
        :param name: the name of the function as a string
        :param f_type: the type of function among (obj, eq, ineq)
        :param args: the function arguments
        """
        self.coefficients = coefficients

        # define the function as a linear combination
        def fun(x_vect):
            """sum_{i=1}^n a_i * x_i

            :param x_vect: design variables
            """
            return inner(coefficients, x_vect)

        # Jacobian is constant
        def jac(_):
            """Constant Jacobian is the matrix of coefficients

            :param _: design variables

            """
            return atleast_2d(coefficients)

        expr = self._generate_expr()
        args = ["x"]

        super(MDOLinearFunction, self).__init__(
            fun, name, f_type=f_type, jac=jac, expr=expr, args=args, dim=1
        )

    def _generate_expr(self):
        """Generate the literal expression of the linear function"""
        expr = ""
        first_non_zero = -1
        for i in range(len(self.coefficients)):
            if self.coefficients[i] != 0.0:
                if first_non_zero == -1:
                    first_non_zero = i
                # sign
                if self.coefficients[i] == -1.0:
                    if i == first_non_zero:
                        expr += "-"  # unary minus
                elif i > 0 and first_non_zero < i:
                    if self.coefficients[i] < 0.0:
                        expr += " - "
                    else:
                        expr += " + "
                # coefficient
                if abs(self.coefficients[i]) != 1.0:
                    expr += str(abs(self.coefficients[i])) + "*"
                # variable
                expr += "x_" + str(i)
        return expr


class MDOFunctionGenerator(with_metaclass(SingleInstancePerAttributeId, object)):
    """The link between MDODisciplines and objective
    functions and constraints is made with MDOFunctionGenerator,
    which generates MDOFunctions from the disciplines.

    It uses closures to generate functions instances from a discipline
    execution.
    """

    def __init__(self, discipline):
        """
        Constructor

        :param discipline: the discipline from which the generator
                           builds functions
        """
        self.discipline = discipline

    def __make_function(self, input_names_list, output_names_list, default_inputs):
        """
        Makes a function from io and reference data
        Combines lambda function and a local def for function pointer issues
        Uses a closure.

        :param input_names_list: the dict keys of the input data
        :param output_names_list: the dict keys of the output data
        :param default_inputs: a default inputs dict to eventually overload
            the discipline's default inputs when evaluating the discipline
        :returns: The generated function
        """

        def func(x_vect):
            """A function which executes a discipline

            :param x_vect: the input vector of the function
            :type x_vect: ndarray
            :returns: the selected outputs of the discipline
            :rtype: ndarray
            """
            for name in input_names_list:
                if name not in self.discipline.default_inputs:
                    raise ValueError(
                        "Discipline "
                        + str(self.discipline.name)
                        + " has no default_input named "
                        + str(name)
                        + ", while input is required"
                        + " by MDOFunction."
                    )
            defaults = self.discipline.default_inputs
            if default_inputs is not None:
                defaults.update(default_inputs)
            data = DataConversion.update_dict_from_array(
                defaults, input_names_list, x_vect
            )
            self.discipline.reset_statuses_for_run()
            computed_values = self.discipline.execute(data)
            values_array = DataConversion.dict_to_array(
                computed_values, output_names_list
            )
            if values_array.size == 1:  # Then the function is scalar
                return values_array[0]
            return values_array

        def func_jac(x_vect):
            """A function which linearizes a discipline

            :param x_vect: the input vector of the function
            :type x_vect: ndarray
            :returns: the selected outputs of the discipline
            :rtype: ndarray
            """
            defaults = self.discipline.default_inputs
            n_dv = len(x_vect)
            data = DataConversion.update_dict_from_array(
                defaults, input_names_list, x_vect
            )
            self.discipline.linearize(data)

            grad_array = []
            for out_name in output_names_list:
                jac_loc = self.discipline.jac[out_name]
                grad_loc = DataConversion.dict_to_array(jac_loc, input_names_list)
                grad_output = hstack(grad_loc)
                if len(grad_output) > n_dv:
                    grad_output = reshape(grad_output, (grad_output.size // n_dv, n_dv))
                grad_array.append(grad_output)
            grad = vstack(grad_array).real
            if grad.shape[0] == 1:
                grad = grad.flatten()
                assert len(x_vect) == len(grad)
            return grad

        default_name = "_".join(output_names_list)
        mdo_func = MDOFunction(
            func,
            name=default_name,
            args=input_names_list,
            outvars=output_names_list,
            jac=func_jac,
        )
        return mdo_func

    def get_function(self, input_names_list, output_names_list, default_inputs=None):
        """Builds a function from a discipline input and output lists.

        :param input_names_list: names of inputs of the disciplines
           to be inputs of the function
        :param output_names_list: names of outputs of the disciplines
            to be returned by the function
        :param default_inputs: a default inputs dict to eventually overload
            the discipline's default inputs when evaluating the discipline
        :returns: the function
        """
        if isinstance(input_names_list, string_types):
            input_names_list = [input_names_list]

        if isinstance(output_names_list, string_types):
            output_names_list = [output_names_list]

        if input_names_list is None:
            input_names_list = self.discipline.get_input_data_names()
        if output_names_list is None:
            output_names_list = self.discipline.get_output_data_names()

        if not self.discipline.is_all_inputs_existing(input_names_list):
            raise ValueError(
                "Some elements of "
                + str(input_names_list)
                + " are not inputs of the discipline, "
                + str(self.discipline.name)
                + " available inputs are:"
                + str(self.discipline.get_input_data_names())
            )

        if not self.discipline.is_all_outputs_existing(output_names_list):
            raise ValueError(
                "Some elements of "
                + str(output_names_list)
                + " are not outputs of the discipline, "
                + str(self.discipline.name)
                + " available outputs are: "
                + ", ".join(self.discipline.get_output_data_names())
                + "."
            )

        # adds inputs and outputs to the list of variables to be
        # differentiated
        self.discipline.add_differentiated_inputs(input_names_list)
        self.discipline.add_differentiated_outputs(output_names_list)

        return self.__make_function(input_names_list, output_names_list, default_inputs)


class NotImplementedCallable(object):
    """
    A not implemented callable object
    """

    def __call__(self, *args, **kwargs):
        raise NotImplementedError("Function is not implemented")



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#        :author: Damien Guenot
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Plot the derivatives of the functions
*************************************
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from matplotlib import pyplot
from numpy import arange, atleast_2d

from gemseo.post.opt_post_processor import OptPostProcessor

standard_library.install_aliases()


from gemseo import LOGGER


class GradientSensitivity(OptPostProcessor):
    """
    The **GradientSensitivity** post processing
    builds histograms of derivatives of objective and constraints

    The plot method considers the derivatives at the last iteration.
    The iteration can be changed in option. The x- and y- figure sizes
    can also be modified in option.
    It is possible either to save the plot, to show the plot or both.
    """

    def _plot(
        self,
        iteration=-1,
        figsize_x=10,
        figsize_y=10,
        save=False,
        show=False,
        file_path="gradient_sensitivity",
        extension="pdf",
    ):
        """
        Plots the GradientSensitivity graph

        :param iteration:  the iteration to plot sensitivities, if negative,
            use optimum
        :type iteration: int
        :param figsize_x: size of figure in horizontal direction (inches)
        :type figsize_x: int
        :param figsize_y: size of figure in vertical direction (inches)
        :type figsize_y: int
        :param show: if True, displays the plot windows
        :type show: bool
        :param save: if True, exports plot to pdf
        :type save: bool
        :param file_path: the base paths of the files to export
        :type file_path: str
        :param extension: file extension
        :type extension: str
        """
        all_funcs = self.opt_problem.get_all_functions_names()

        if iteration == -1:
            x_ref = self.opt_problem.solution.x_opt
        else:
            x_ref = self.opt_problem.database.get_x_by_iter(iteration)
        grad_dict = {}
        for func in all_funcs:
            grad = self.database.get_f_of_x("@" + func, x_ref)
            if grad is not None:
                if len(grad.shape) == 1:
                    grad_dict[func] = grad
                else:
                    n_f, _ = grad.shape
                    for i in range(n_f):
                        grad_dict[func + "_" + str(i)] = grad[i, :]

        fig = self.__generate_subplots(x_ref, grad_dict, figsize_x, figsize_y)
        self._save_and_show(
            fig, save=save, show=show, file_path=file_path, extension=extension
        )

    @staticmethod
    def __generate_subplots(x_ref, grad_dict, figsize_x=10, figsize_y=10):
        """
        Generates the gradient sub plots from the data

        :param x_ref: reference value for x
        :param grad_dict : dict of gradients to plot
        :param figsize_x : size of figure in horizontal direction (inches)
        :param figsize_y : size of figure in vertical direction (inches)
        """
        n_funcs = len(grad_dict)
        if n_funcs == 0:
            raise ValueError("No gradients to plot at current iteration !")
        nrows = n_funcs // 2
        if 2 * nrows < n_funcs:
            nrows += 1
        ncols = 2
        fig, axes = pyplot.subplots(
            nrows=nrows,
            ncols=2,
            sharex=True,
            sharey=False,
            figsize=(figsize_x, figsize_y),
        )
        i = 0
        j = -1

        axes = atleast_2d(axes)
        n_subplots = len(axes) * len(axes[0])
        abscissa = arange(len(x_ref))
        x_labels = [r"$x_{" + str(x_id) + "}$" for x_id in abscissa]
        for func, grad in sorted(grad_dict.items()):
            j += 1
            if j == ncols:
                j = 0
                i += 1
            axe = axes[i][j]
            axe.bar(abscissa, grad, color="blue", align="center")
            axe.set_title(func)
            axe.set_xticklabels(x_labels, fontsize=14)
            axe.set_xticks(abscissa)
            # Update y labels spacing
            vis_labels = [
                label for label in axe.get_yticklabels() if label.get_visible() is True
            ]
            pyplot.setp(vis_labels[::2], visible=False)

        if len(grad_dict) < n_subplots:
            # xlabel must be written with the same fontsize on the 2 columns
            j += 1
            #             if j == ncols: Seems impossible to reach
            #                 j = 0
            #                 i += 1
            axe = axes[i][j]
            axe.set_xticklabels(x_labels, fontsize=14)
            axe.set_xticks(abscissa)

        fig.suptitle(
            "Derivatives of objective and constraints"
            + " with respect to design variables",
            fontsize=14,
        )
        return fig



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Provide routines for XDSM and tikz."""

from __future__ import absolute_import, division, unicode_literals

import os
from pathlib import Path

from future import standard_library

from gemseo.third_party.pyxdsm.XDSM import XDSM
from gemseo.utils.py23_compat import string_types

standard_library.install_aliases()


def __add_processes(xdsm, workflow, prev=None):
    """Create the pyXDSM processes from a given workflow
    in a recursive way.

    :param XDSM xdsm: XDSM object
    :param list workflow: workflow component of the dictionary
        storing the XDSM
    :param str prev: name of the previous node
    """
    systems = []
    last_node = None
    for idx, system in enumerate(workflow):
        if isinstance(system, string_types):
            # system is a node
            systems.append(system)
            last_node = system
        elif isinstance(system, list):
            # system is a group of nodes
            if isinstance(system[0], string_types):
                # system[0] is a node
                if last_node is not None:
                    xdsm.add_process([last_node, system[0]], arrow=True)
                __add_processes(xdsm, system, workflow[idx - 1])
            elif isinstance(system[0], dict):
                # system[0] is a parallel sequence
                for sub_sys in system[0]["parallel"]:
                    # add a sub-process for each parallel element
                    sub_workflow = [last_node, sub_sys]
                    __add_processes(xdsm, sub_workflow, last_node)
    if prev is not None:
        systems.append(prev)
    xdsm.add_process(systems, arrow=True)


def __get_numbers(numbers, nodes, current=0, end=0, following=1, prev_node=None):
    """Give number to the different nodes in a recursive way.

    :param numbers: dictionary
    :param list nodes: workflow component of the dictionary
        storing the XDSM
    :param int current: current step index
    :param int end: end-loop step index
    :param int following: following step index
    """
    for node in nodes:
        if isinstance(node, string_types):
            current = following
            following = current + 1
            end = current
            current_l = [current]
            if node in list(numbers.keys()):
                current_l = numbers[node]["current"] + current_l
            numbers[node] = {"current": current_l, "next": following, "end": end}
            prev_node = node
        elif isinstance(node, list):
            following = __get_numbers(numbers, node, current, end, following, prev_node)
            numbers[prev_node]["end"] = following
            following += 1
        elif isinstance(node, dict):
            for sub_nodes in node["parallel"]:
                if not isinstance(sub_nodes, list):
                    following = __get_numbers(
                        numbers, [sub_nodes], current, end, following
                    )
                else:
                    following = __get_numbers(
                        numbers, sub_nodes, current, end, following
                    )
                following -= 1
            following += 1
    return following


def __add_nodes(xdsm, numbers, nodes):
    """Add the different nodes, called 'systems', in the XDSM."""
    for node in nodes:
        name = ",".join([str(current) for current in numbers[node["id"]]["current"]])

        name_1 = name + ",{}-{}:".format(
            str(numbers[node["id"]]["end"]), str(numbers[node["id"]]["next"])
        )
        name_2 = name + ":"

        if node["type"] == "optimization":
            node_type = "Optimization"
            name = name_1
        elif node["type"] == "lp_optimization":
            node_type = "LP_Optimization"
            name = name_1
        if node["type"] == "doe":
            node_type = "DOE"
            name = name_1
        elif node["type"] == "mda":
            node_type = "MDA"
            if node["name"] == "MDAChain":
                name = name_2
            else:
                name = name_1
        elif node["type"] == "mdo":
            node_type = "MDO"
            name = name_2
        elif node["type"] == "analysis":
            node_type = "Analysis"
            name = name_2
        elif node["type"] == "function":
            node_type = "Function"
            name = name_2
        elif node["type"] == "metamodel":
            node_type = "Metamodel"
            name = name_2

        node_replaced = node["name"]
        escaped_characters = ["_", "$", "&", "{", "}", "%"]
        for char in escaped_characters:
            node_replaced = node_replaced.replace(char, r"\{}".format(char))
        name = name + node_replaced
        xdsm.add_system(node["id"], node_type, r"\text{" + name + "}")


def __add_edges(xdsm, edges):
    """Add the edges called connections, inputs, outputs to the XDSM."""
    for edge in edges:
        old_names = edge["name"].split(",")
        names = []

        for name in old_names:
            name = name.replace("(0)", "{(0)}")
            if "*" in name:
                if "_" in name:
                    name = name.replace("^*", "")
                    name = name.replace("_", "^{*}_{")
            elif "(0)" in name:
                if "_" in name:
                    name = name.replace("(0)", "")
                    name = name.replace("_", "^{(0)}_{")
                else:
                    name = name.replace("(0)", "{(0)}")
            else:
                name = name.replace("_", "_{")
            name += "}" * name.count("_")
            names.append(name)

        names = ", ".join(names)

        if edge["to"] == "_U_":
            xdsm.add_output(edge["from"], r"" + names + "", side="left")
        elif edge["from"] != "_U_":
            xdsm.connect(edge["from"], edge["to"], r"" + names + "")
        elif edge["from"] == "_U_":
            xdsm.add_input("Opt", r"" + names + "")


def xdsm_dict2tex(dict_xdsm, out_dir, out_filename="xdsm", scenario="root", quiet=True):
    """Convert a dictionary representation of a XDSM
    into a pdf or tikz representation.

    :param dict xdsm: XDSM dictionary representation
    :param str out_dir: output directory
    :param str out_filename: output file name (default: 'xdsm')
    :param str scenario: scenario name (default: 'root')
    :param bool quiet: set to True to suppress output from pdflatex.
    """
    workflow = dict_xdsm[scenario]["workflow"][1]

    numbers = {}
    __get_numbers(numbers, workflow)

    xdsm = XDSM()

    __add_nodes(xdsm, numbers, dict_xdsm[scenario]["nodes"])
    __add_edges(xdsm, dict_xdsm[scenario]["edges"])
    __add_processes(xdsm, workflow)

    out_texfile = os.path.join(out_dir, out_filename)
    out_xdsm = os.path.join(out_dir, out_filename) + ".pdf"
    xdsm.write(out_texfile, quiet=quiet)

    if not Path(out_xdsm).exists():
        raise RuntimeError(
            "Something went wrong during the Latex compilation,"
            " as xdsm.pdf has not been generated. Please have a look at the Latex log file"
            " to investigate the root cause of the error."
        )

    # Build XDSM for sub-scenarios
    if scenario == "root":
        subscenarios = [key for key in dict_xdsm.keys() if key.startswith("scn-")]
        for subscenario in subscenarios:
            xdsm_dict2tex(
                dict_xdsm,
                out_dir,
                out_filename=out_filename + "_" + subscenario,
                scenario=subscenario,
                quiet=quiet,
            )
    return numbers



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Matthias De Lozzo, Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Classification model
====================

The :mod:`~gemseo.mlearning.classification.classification` module
implements classification algorithms, whose goal is to find relationships
between input data and output classes.

Classification algorithms provide methods for predicting classes of new input
data, as well as predicting the probabilities of belonging to each of the
classes wherever possible.

This concept is implemented through the :class:`.MLClassificationAlgo` class
which inherits from the :class:`.MLSupervisedAlgo` class.
"""
from __future__ import absolute_import, division, unicode_literals

from future import standard_library
from numpy import unique, zeros

from gemseo.mlearning.core.supervised import MLSupervisedAlgo

standard_library.install_aliases()


class MLClassificationAlgo(MLSupervisedAlgo):
    """Classification Algorithm.

    Inheriting classes should implement the :meth:`!MLSupervisedAlgo._fit` and
    :meth:`!MLClassificationAlgo._predict` methods, and
    :meth:`!MLClassificationAlgo._predict_proba_soft` method if possible.
    """

    def __init__(
        self, data, transformer=None, input_names=None, output_names=None, **parameters
    ):
        """Constructor.

        :param Dataset data: learning dataset.
        :param transformer: transformation strategy for data groups.
            If None, do not scale data. Default: None.
        :type transformer: dict(str)
        :param input_names: names of the input variables.
        :type input_names: list(str)
        :param output_names: names of the output variables.
        :type output_names: list(str)
        :param parameters: algorithm parameters.
        """
        super(MLClassificationAlgo, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            **parameters
        )
        self.n_classes = None

    def learn(self, samples=None):
        """Train machine learning algorithm on learning set, possibly filtered
        using the given parameters. Determine the number of classes.
        :param list(int) samples: indices of training samples.
        """
        output_data = self.learning_set.get_data_by_names(self.output_names, False)
        self.n_classes = unique(output_data).shape[0]
        super(MLClassificationAlgo, self).learn(samples)

    @MLSupervisedAlgo.DataFormatters.format_input_output
    def predict_proba(self, input_data, hard=True):
        """Predict probability of belonging to each class.

        :param input_data: input data (n_inputs,) or (n_samples, n_inputs).
        :type input_data: dict(ndarray) or ndarray
        :param bool hard: indicator for hard or soft classification.
            Default: True.
        :return: probabilities of belonging to each class
            (n_outputs, n_classes) or (n_samples, n_outputs, n_classes).
        :rtype: dict(ndarray) or ndarray
        """
        return self._predict_proba(input_data, hard)

    def _predict_proba(self, input_data, hard=True):
        """Predict probability of belonging to each class.

        :param ndarray input_data: input data (n_samples, n_inputs).
        :param bool hard: indicator for hard or soft classification.
            Default: True.
        :return: probabilities of belonging to each class
            (n_samples, n_outputs, n_classes). For a given sample and output
            variable, the sum of probabilities is one.
        :rtype: ndarray
        """
        if hard:
            probas = self._predict_proba_hard(input_data)
        else:
            probas = self._predict_proba_soft(input_data)
        return probas

    def _predict_proba_hard(self, input_data):
        """Create class indicator of input data.

        :param ndarray input_data: input data (n_samples, n_inputs).
        :return: probabilities of belonging to each class
            (n_samples, n_outputs, n_classes). For a given sample and output
            variable, the sum of probabilities is one.
        :rtype: ndarray
        """
        n_samples = input_data.shape[0]
        prediction = self._predict(input_data).astype(int)
        n_outputs = prediction.shape[1]
        probas = zeros((n_samples, self.n_classes, n_outputs))
        for n_sample in range(prediction.shape[0]):
            for n_output in range(n_outputs):
                probas[n_sample, prediction[n_sample, n_output], n_output] = 1
        return probas

    def _predict_proba_soft(self, input_data):
        """Predict probability of belonging to each class.

        :param ndarray input_data: input data (n_samples, n_inputs).
        :return: probabilities of belonging to each class
            (n_samples, outputs, n_classes). For a given sample and output
            variable, the sum of probabilities is one.
        :rtype: ndarray
        """
        raise NotImplementedError

    def _get_objects_to_save(self):
        """Get objects to save.
        :return: objects to save.
        :rtype: dict
        """
        objects = super(MLClassificationAlgo, self)._get_objects_to_save()
        objects["n_classes"] = self.n_classes
        return objects



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A Grammar based on JSON schema
******************************
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import os

from future import standard_library
from numpy import generic, ndarray, zeros

from gemseo import LOGGER
from gemseo.core.grammar import AbstractGrammar, InvalidDataException
from gemseo.third_party.genson_generator import Schema
from gemseo.utils.py23_compat import JsonSchemaException, compile_schema

standard_library.install_aliases()


class JSONGrammar(AbstractGrammar):
    """A grammar subclass that stores the input or output data types
    and structure a MDODiscipline using a JSON format
    It is able to check the inputs and outputs against a JSON schema.
    """

    PROPERTIES_FIELD = "properties"
    DEFAULTS_FIELD = "defaults"
    REQUIRED_FIELD = "required"
    TYPE_FIELD = "type"
    OBJECT_FIELD = "object"

    def __init__(
        self,
        name,
        schema_file=None,
        schema=None,
        grammar_type=AbstractGrammar.INPUT_GRAMMAR,
    ):
        """
        Constructor
        :param name : grammar name
        :param schema : a genson Schema  to initialize self
        :param schema_file : the json schema file
        :param grammar_type : the type of grammar : input or output
        """
        super(JSONGrammar, self).__init__()
        self.schema = None
        self._properties = None
        self._validator = None
        self.data = None
        self.name = name
        self.grammar_type = grammar_type
        self._init_schema()
        if schema is not None:
            self.schema.add_schema(schema)
            self._update_properties()
        if schema_file is not None:
            self.init_from_schema_file(schema_file)
        if schema is None and schema_file is None:
            self.initialize_from_base_dict(typical_data_dict={})

    def _init_schema(self):
        """Initializes the genson schema"""
        self.schema = Schema(merge_arrays=True, exclude_at_merge=["name", "id"])

    def _update_properties(self):
        """Updates the properties after a change in the schema"""
        schema_dict = self.schema.to_dict()
        self._properties = schema_dict.get(self.PROPERTIES_FIELD)
        if self._properties is None:
            self._properties = {}

        self._validator = None

    @property
    def properties(self):
        """Accessor for the properties of the schema"""
        return self._properties

    def clear(self):
        """Clears the data to produce an empty grammar"""
        self._init_schema()
        self._update_properties()
        self._validator = None

    def _init_validator(self):
        """Initializes the validator according to self"""
        schema_dict = self.schema.to_dict()
        schema_dict.pop("id", None)
        self._validator = compile_schema(schema_dict)

    @staticmethod
    def cast_array_to_list(data_dict):
        """
        Casts the numpy arrays in data_dict to lists

        :param data_dict : the data dictionary
        :returns: The dict with casted arrays
        """
        out_d = data_dict.copy()
        for key, value in data_dict.items():
            if isinstance(value, (ndarray, generic)):
                out_d[key] = value.real.tolist()
            elif isinstance(value, dict):
                out_d[key] = JSONGrammar.cast_array_to_list(value)
        return out_d

    def load_data(self, data_dict, raise_exception=True):
        """Loads the data dictionary in the grammar
        and checks it against json schema

        :param data_dict: the input data
        :param raise_exception: if False, no exception is raised
            when data is invalid (Default value = True)
        """
        if not isinstance(data_dict, dict):
            raise InvalidDataException(
                "Input data must be a python dictionary, got "
                + str(type(data_dict))
                + " instead."
            )
        error_exist = False
        data_to_check = self.cast_array_to_list(data_dict)

        if self._validator is None:
            self._init_validator()

        try:
            self._validator(data_to_check)
        except JsonSchemaException as error:
            error_exist = True
            msg = "Invalid data in : " + str(self.name)

            if error.args[0].startswith("data must contain"):
                # Error messages are not clear enough when missing properties
                # All keys are put in the message
                diff = set(set(self.get_data_names()) - set(data_dict.keys()))
                if len(diff) > 0:
                    msg += "\nMissing mandatory properties: " + str(list(diff))
                else:
                    msg += "\n', error : " + str(error.args[0])
            else:
                msg += "\n', error : " + str(error.args[0])
            LOGGER.error(msg)

        if error_exist and raise_exception:
            raise InvalidDataException("Invalid data from grammar " + str(self.name))

        # Check a copy to keep types and arrays but store initial dict for
        # complex
        self.data = data_dict
        # Add defaults
        for key, value in data_to_check.items():
            k_utf = key
            self.data.setdefault(k_utf, value)
        return self.data

    def init_from_schema_file(self, schema_file="input.json"):
        """Initializes grammar from

        :param schema_file: path to the schema input file
            (Default value = "input.json")
        """
        if not os.path.exists(schema_file):
            raise ValueError(
                "Try to initialize grammar with not existing file : " + str(schema_file)
            )

        # Refs are not supported any more for simplification
        # curr_folder = os.path.dirname(schema_file)
        # registry = providers.FilesystemProvider(curr_folder)

        with open(schema_file, "r") as in_file:
            json_content = json.loads(in_file.read())
            # Refs not supported any more for simplification
            # Resolve #ref tags in json
            #             try:
            #              resolved_json = resolve(json_content, '#/properties', registry)
            #             except ValueError as err_json:
            #                 LOGGER.error(err_json)
            #                 raise Exception(
            #                     "Cannot resolve referenced json properties of file : " +
            #                     str(schema_file))
            self.schema.add_schema(json_content)
            self._update_properties()

    def __str__(self):
        msg = "Grammar named :" + str(self.name)
        msg += ", schema = " + self.schema.to_json()
        return msg

    def initialize_from_data_names(
        self, data_names, schema_file=None, write_schema=False
    ):
        """Initializes a JSONGrammar from a list of data.
        All data of the grammar will be set as arrays

        :param data_names: a data names list
        :param schema_file: the output json file path. If None : input.json or
            output.json depending on grammar type.
            (Default value = None)
        :param write_schema: if True, writes the schema files
            (Default value = False)
        """
        data = zeros(1)
        typical_data_dict = {k: data for k in data_names}
        self.initialize_from_base_dict(typical_data_dict, schema_file, write_schema)

    def initialize_from_base_dict(
        self,
        typical_data_dict,
        schema_file=None,
        write_schema=False,
        description_dict=None,
    ):
        """Initialize a json grammar with types and names from a
        typical data entry.
        The keys of the typical_data_dict will be the names of the
        data in the grammar.
        The types of the values of the typical_data_dict will be converted
        to JSON Schema types and define the properties of the JSON Schema.

        :param typical_data_dict: a data dictionary with keys as data names
            and values as a typical value for this data
        :param schema_file: the output json file path. If None : input.json or
            output.json depending on grammar type.
            (Default value = None)
        :param write_schema: if True, writes the schema files
            (Default value = False)
        :param description_dict: dictionary of descriptions,
             {name:meaning} structure
        """
        # Convert arrays to list as for check
        list_data_dict = self.cast_array_to_list(typical_data_dict)
        #         if PY2:
        #             list_data_dict = self.cast_str_val_ascii(list_data_dict)
        self.schema.add_object(list_data_dict, description_dict)

        if write_schema:
            if schema_file is None:
                schema_file = self.name + ".json"
            with open(schema_file, "w") as outf:
                outf.write(self.schema.to_json())
        self._update_properties()

    #     @staticmethod
    #     def cast_str_val_ascii(data_dict):
    #         """
    #         Casts the values of dict to ascii
    #
    #         :param: input data dict
    #         :returns: ourput data dict
    #         """
    #         out_d = data_dict.copy()
    #         for key, value in data_dict.items():
    #             if isinstance(value, string_types):
    #                 out_d[key] = value.encode('ascii', 'ignore')
    #             elif isinstance(value, dict):
    #                 out_d[key] = JSONGrammar.cast_str_val_ascii(value)
    #         return out_d

    def add_description(self, description_dict):
        """
        Add a description to the properties

        :param description_dict: dictionary of descriptions,
             {name:meaning} structure
        """

        descr_filtered = {
            k: v for k, v in description_dict.items() if k in self._properties
        }
        self.schema.add_object({}, descr_filtered)

    def get_data_names(self):
        """Returns the list of data names

        :returns: the data names, as a dict keys set
        """
        return self._properties.keys()

    def is_data_name_existing(self, data_name):
        """Checks if data_name is present in grammar

        :param data_name: the data name
        :returns: True if data is in grammar
        """
        return data_name in self._properties

    def is_all_data_names_existing(self, data_names):
        """Checks if data_names are present in grammar

        :param data_names: the data names list
        :returns: True if all data are in grammar
        """
        exists = self.is_data_name_existing
        for data_name in data_names:
            if not exists(data_name):
                return False
        return True

    def update_from(self, input_grammar):
        """Adds properties coming from another grammar

        :param input_grammar: the grammar to take inputs from
        """
        if not isinstance(input_grammar, JSONGrammar):
            msg = "Cannot update grammar " "{} of type {} with {} of type {} ".format(
                self.name,
                type(self).__name__,
                input_grammar.name,
                type(input_grammar).__name__,
            )
            raise ValueError(msg)

        schema_dct = input_grammar.schema.to_dict()
        self.schema.add_schema(schema_dct)
        self._update_properties()

    def update_from_if_not_in(self, input_grammar, exclude_grammar):
        """Adds objects coming from input_grammar if they are not in
        exclude_grammar

        :param input_grammar: the grammar to take inputs from
        :param exclude_grammar: exclusion grammar
        """
        if isinstance(input_grammar, JSONGrammar) and isinstance(
            exclude_grammar, JSONGrammar
        ):
            in_schema_dct = input_grammar.schema.to_dict()
            in_schema_prop = in_schema_dct.get(self.PROPERTIES_FIELD, {})
            in_required = in_schema_dct.get(self.REQUIRED_FIELD, [])
            ex_schema_dct = exclude_grammar.schema.to_dict()
            ex_schema_prop = ex_schema_dct.get(self.PROPERTIES_FIELD, {})
            merged_required = []
            merged_prop = {}
            for prop_name, prop_schema in in_schema_prop.items():
                if prop_name not in ex_schema_prop:
                    merged_prop[prop_name] = prop_schema
                    if prop_name in in_required:
                        merged_required.append(prop_name)

            merged_schema = {
                self.TYPE_FIELD: self.OBJECT_FIELD,
                self.PROPERTIES_FIELD: merged_prop,
                self.REQUIRED_FIELD: merged_required,
            }

            self.schema.add_schema(merged_schema)
            self._update_properties()
        else:

            msg = "Cannot update grammar " + str(self.name)
            msg += " of type JSONGrammar with " + str(input_grammar.name)
            msg += " of type " + str(type(input_grammar).__name__)
            msg += " and " + str(exclude_grammar.name)
            msg += ", of type " + str(type(exclude_grammar).__name__)
            raise TypeError(msg)

    def restrict_to(self, data_names):
        """Restricts the grammar to a sublist of data names

        :param data_names: the names of the data to restrict the grammar to
        """

        self_schema_dct = self.schema.to_dict()
        self_schema_prop = self_schema_dct.get(self.PROPERTIES_FIELD, {})
        self_required = self_schema_dct.get(self.REQUIRED_FIELD, [])
        for prop_name in list(self_schema_prop.keys()):
            if prop_name not in data_names:
                del self_schema_prop[prop_name]
                if prop_name in self_required:
                    self_required.remove(prop_name)

        self_schema = {
            self.TYPE_FIELD: self.OBJECT_FIELD,
            self.PROPERTIES_FIELD: self_schema_prop,
            self.REQUIRED_FIELD: self_required,
        }

        self._init_schema()
        self.schema.add_schema(self_schema)
        self._update_properties()

    def remove_item(self, item_name):
        """Removes an item from the grammar

        :param item_name: the item name to be removed

        """
        if not self.is_data_name_existing(item_name):
            raise ValueError("Item " + str(item_name) + " not in grammar " + self.name)
        self_dict = self.schema.to_dict()
        del self_dict[self.PROPERTIES_FIELD][item_name]
        self_required = self_dict.get(self.REQUIRED_FIELD, [])
        if item_name in self_required:
            self_required.remove(item_name)
        self._init_schema()
        self.schema.add_schema(self_dict)
        self._update_properties()

    def set_item_value(self, item_name, item_value):
        """
        Sets the value of an item

        :param item_name: the item name to be modified
        :param item_value: value of the item
        """
        if not self.is_data_name_existing(item_name):
            raise ValueError("Item " + str(item_name) + " not in grammar " + self.name)
        self_dict = self.schema.to_dict()
        self_dict[self.PROPERTIES_FIELD][item_name] = item_value
        self._init_schema()
        self.schema.add_schema(self_dict)
        self._update_properties()

    def __getstate__(self):
        """
        Used by pickle to define what to serialize

        :returns: the dict to serialize
        """
        out_d = {}
        out_d.update(self.__dict__)
        out_d.pop("_validator")
        return out_d

    def __setstate__(self, data_dict):
        """
        Used by pickle to define what to deserialize

        :param data_dict : update self dict from data_dict to deserialize
        """
        self.__dict__.update(data_dict)
        self._update_properties()
